#-------------------------------------------------------------------------------
# Name:        manage_vc
# Purpose:     Manage one or more virtual computers
#
# Author:      Stephen Rondeau
#
# Created:     5/16/2022
#-------------------------------------------------------------------------------

from tkinter import *
from tkinter import ttk 
from tkinter import messagebox, simpledialog

import os
import sys
import time
import re
from configparser import ConfigParser
import platform
##import tempfile
import getpass
from datetime import datetime
import cmd
import subprocess, shlex
import signal
from idlelib.tooltip import Hovertip
from functools import partial
  
if platform.system() == "Darwin":
  from tkmacosx import *

import pexpect
from pexpect import popen_spawn

#------------------------------------------------------------------------------
# The main thing that a user will do is start the VM and connect to its
# display. Thereafter, they will use the RDP client to interact with the
# VM, and can shutdown the VM via the guest OS. 
# 
# In the early years of the VCL, students would often install their
# own OS into a pre-defined VM, and hence would require more control
# over the VM. This hasn't been done at all in the last several years,
# since faculty prefer to have the guest OS installed and updated at
# the start of the course. It also provides an ability for SET Lab staff
# to install software (like VirtualBox Guest Additions) to make the
# user experience better (better mouse tracking, syncing VM time to
# host time, etc.) or to prepare the VM for course needs based on
# what the instructor requested.
#
# Managing a VM beyond just starting a VM and connecting to its display
# is more advanced, and is often used in system admin courses
# or to fix initial VM misconfigurations. Note that the original impetus 
# for manage_vc was to remove SET Lab staff from needing to change VM
# settings for students or to start VMs after either the host or they
# shut them down. Having one or two staff doing that is not
# scalable, is disruptive for students (wait until staff get a chance
# to fix it) and would have limited the utility of the VCL.
#------------------------------------------------------------------------------
def create_gui():
  global tk_root

  tk_root = Tk()
  tk_root.title(mvv)

  tk_root.style = ttk.Style()
  tk_root.iconphoto(True, PhotoImage(file = 'manage_vc.png'))
  
  set_main_geometry()

  tk_root.protocol("WM_DELETE_WINDOW", on_close)

  # Make main window invisible so pre-VM window dialog boxes can get focus.
  tk_root.attributes('-alpha', 0.0)
  #tk_root.withdraw() # Hide main window prior to extracting VCL info


#------------------------------------------------------------------------------
def set_main_geometry():
  global tk_root
  # get the screen dimension
  screen_width = tk_root.winfo_screenwidth()
  screen_height = tk_root.winfo_screenheight()

  center_x = int(screen_width/2 - main_window_width / 2)
  center_y = int(screen_height/2 - main_window_height / 2)
  tk_root.geometry(str(main_window_width) + 'x' + str(main_window_height) + '+' 
                   + str(center_x) + '+' + str(center_y))

#------------------------------------------------------------------------------
def populate_gui():
  global tk_root, vms, main_window_width, main_window_width_adj
  tk_root.title(mvv + ': ' + vcl_fn_user)

  vms = ttk.Notebook(tk_root)
  vms.pack(expand=True)

  create_tabs()
  
  #-------------------------------------------------------------------------
  # If there are more than four VMs, provide enough width for them all.
  #-------------------------------------------------------------------------
  if vmi+1 > 4:
    ##  main_window_width = 20 + (font_width*8)*(vmi+1)
    main_window_width = 20 + ((font_width+1)*max_names_len) + main_window_width_adj
    if platform.system() == "Darwin":
      set_main_geometry()

  # Make main window visible
  tk_root.attributes('-alpha', 1.0)
  #tk_root.deiconify()

def create_tabs():
  global vm_tab, vmi

  vm_tab = [None for _ in range(vmi+1)]
  for i in range(vmi+1):
    create_vm_tab(i)


def create_vm_tab(vi):
  global vm_tab, vms

  vm_tab[vi] = ttk.Frame(vms, width=max_width+20, height=max_height-20)
  vm_tab[vi].pack(fill='both', expand=True)

  vms.add(vm_tab[vi], text=tab_name[vi])

  # Create tab item components
  create_tab(vi)

def create_tab(vi):
  global connect_button, connect_vrdp, connect_grdp, connect_edit
  global screenshot_button, screenshot_list, screenshot_combo, screenshot_secs_label
  global set_nic_button, nic_combo, nic_value

  global start_vc_button, cae_button, power_off_button
  global show_state_vc_button

  global eject_cddvd_button, add_cd_button, show_cd_button
  global add_cd_combo, add_cd_value

  global reveal_pw_button, paste_pw_button, guest_user_label, guest_user, guest_user_text

  global reveal_ip_button
  global paste_vc_button
  global check_ssh_button

  global info_list
  global il_frame
  global il_scrollbar
  global do_cmd_button, cmd_combo

  global rdp_title
  global copy_host_and_port_button

  global top_pos

  global check_rdp, check_edit

  global max_names_len

  tab = vm_tab[vi]
  curr_x_pos = 10

  #----------------------------------------------------------------------
  # Since names can be long and won't show on some controls, we provide a
  # short name for any VM name longer than 8 characters. The short name
  # is six characters followed by an ellipsis. That means that we have
  # a max width for any control that uses a VM name, and can provide for
  # that in the control's width.
  #----------------------------------------------------------------------
  if len(vm_name[vi]) > 8:
    short_vm_name = vm_name[vi][:6] + "..."
  else:
    short_vm_name = vm_name[vi]

  max_names_len = max_names_len + len(short_vm_name)

  if platform.system() == "Windows":
    width_adj = 1
  else:
    width_adj = 5

  #---------------------------------------------------------------------
  # Connect
  #---------------------------------------------------------------------
  name = "Connect " + short_vm_name
  ##connect_button[vi] = Button(tab, text=name, command=on_connect, fg="green", width=len(name)-width_adj)
  width = get_button_width(name, -2, 0, 0)

  connect_button[vi] = Button(tab, text=name, command=on_connect, bg="white", fg="green", highlightbackground="#e3e3e3", width=width)

  connect_button[vi].place(x=curr_x_pos, y=top_pos+1, height=font_height)
  Hovertip(connect_button[vi], "Connect " + vm_name[vi])

  #---------------------------------------------------------------------
  # If we don't have a VRDP address, rely on the guest OS to provide RDP
  # and allow editing of the values detected or defaulted. Otherwise,
  # prefer using VRDP and don't allow editing by default.
  #---------------------------------------------------------------------
  check_edit[vi] = IntVar(tk_root, 0)
  check_rdp[vi] = IntVar(tk_root, 1)

  if vrdp_addr[vi] == "":
    check_rdp[vi].set(1)
  else:
    check_rdp[vi].set(2)

  #----------------------------------------------------------------------
  # Edit RDP
  #----------------------------------------------------------------------
  connect_edit[vi] = ttk.Checkbutton(tab, variable = check_edit[vi], 
                                     onvalue=1, offvalue=0,
                                     text = "Edit", width=4,
                                     command=on_edit)
  curr_x_pos = (len(name)+2)*font_width
  connect_edit[vi].place(x=curr_x_pos, y=top_pos + 1)

  line_width = curr_x_pos

  #----------------------------------------------------------------------
  # Guest RDP
  #----------------------------------------------------------------------
  name = "Guest RDP"
  ##connect_grdp[vi] = Radiobutton(tab, text=name, variable=check_rdp[vi], value=1, command=on_rdp)
  connect_grdp[vi] = ttk.Radiobutton(tab, text=name, variable=check_rdp[vi], value=1, command=on_rdp)
  curr_x_pos = curr_x_pos + 6*font_width
  connect_grdp[vi].place(x=curr_x_pos, y=top_pos-1, width=(font_width+2)*len(name)+2)

  line_width = line_width + 6*font_width + (font_width+2)*len(name) + 2

  #----------------------------------------------------------------------
  # VRDP
  #----------------------------------------------------------------------
  name = "VRDP"
  ##connect_vrdp[vi] = Radiobutton(tab, text=name, variable=check_rdp[vi], value=2, command=on_rdp)
  connect_vrdp[vi] = ttk.Radiobutton(tab, text=name, variable=check_rdp[vi], value=2, command=on_rdp)
  connect_vrdp[vi].place(x=curr_x_pos, y=top_pos + font_height, width=(font_width+6)*len(name)+2)

  #----------------------------------------------------------------------
  # Start VC
  #----------------------------------------------------------------------
  name = "Start " + short_vm_name
  ## start_vc_button[vi] = Button(tab, text=name, command=on_start_vc, fg="green",  width=len(name)-width_adj)
  width = get_button_width(name, -2, 0, 0)

  start_vc_button[vi] = Button(tab, text=name, command=on_start_vc, bg="green", fg="white",  width=width)

  start_vc_x_pos = line_width
  start_vc_button[vi].place(x=start_vc_x_pos, y=top_pos, height=font_height)
  Hovertip(start_vc_button[vi], "Start " + vm_name[vi])

  old_name = name

  #----------------------------------------------------------------------
  # Ctrl-Alt-End
  #----------------------------------------------------------------------
  name = "C-A-End"
  ## cae_button[vi] = Button(tab, text=name, command=on_send_cae, fg="blue", width=len(name)-2)
  width = get_button_width(name, 0, 4, 0)

  cae_button[vi] = Button(tab, text=name, command=on_send_cae, bg="white", fg="green", highlightbackground="#e3e3e3", width=width)

  curr_x_pos = start_vc_x_pos + (len(old_name))*font_width
  cae_button[vi].place(x=curr_x_pos, y=top_pos, height=font_height)
  Hovertip(cae_button[vi], "Send Ctrl-Alt-End (aka Ctrl-Alt-Del) to VM")

  old_name = name

  line_width = curr_x_pos

  #----------------------------------------------------------------------
  # Power Off VC
  #----------------------------------------------------------------------
  name = "Power Off " + short_vm_name
  ##power_off_button[vi] = Button(tab, text=name, command=on_power_off, fg="red",  width=len(name)-width_adj)
  width = get_button_width(name, -2, 0, 0)

  power_off_button[vi] = Button(tab, text=name, command=on_power_off, bg="red", fg="white",  width=width)

  curr_x_pos = line_width + 10 + (len(old_name)+1)*font_width
  power_off_button[vi].place(x=curr_x_pos, y=top_pos, height=font_height)
  Hovertip(power_off_button[vi], "Power Off " + vm_name[vi])

  old_name = name
  line_width = curr_x_pos

  #----------------------------------------------------------------------
  # Show State
  #----------------------------------------------------------------------
  name = "Show State"
  ##show_state_vc_button[vi] = Button(tab, text=name, command=on_show_state_vc,  bg="#d3d3d3", fg="black", width=len(name)-width_adj)
  width = get_button_width(name, -2, 0, 0)

  show_state_vc_button[vi] = Button(tab, text=name, command=on_show_state_vc,  bg="white", fg="black",highlightbackground="#e3e3e3",  width=width)
  curr_x_pos = start_vc_x_pos
  old_name = name

  show_state_vc_button[vi].place(x=curr_x_pos, y=top_pos+font_height+3, height=font_height)
  line_width = curr_x_pos + len(name)*font_width

  #----------------------------------------------------------------------
  # Set NIC
  #----------------------------------------------------------------------
  name = "Set NIC"
  ##set_nic_button[vi] = Button(tab, text=name, command=on_set_nic,  bg="#E5E4E2", fg="black", width=len(name)-3)
  width = get_button_width(name, -1, 0, 0)

  set_nic_button[vi] = Button(tab, text=name, command=on_set_nic,  bg="white", fg="red", highlightbackground="#e3e3e3", width=width)
  curr_x_pos = line_width

  set_nic_button[vi].place(x=curr_x_pos, y=top_pos+font_height+3, height=font_height)

  old_name = name
  line_width = line_width + 10 + len(name)*font_width

  nic_combo[vi] = ttk.Combobox(tab)
  curr_x_pos = line_width + 2
  ##nic_combo[vi].place(x=curr_x_pos, y=top_pos+font_height+5, height=font_height)
  nic_combo[vi].place(x=curr_x_pos, y=top_pos+font_height+5)

  nic_combo[vi]['values'] = ("nic1 nat", "nic1 intnet " + vcl_fn_user + "_intnet", "nic1 hostonly")

  #----------------------------------------------------------------------
  # Screenshot
  #----------------------------------------------------------------------
  name = "Screenshot every"
  ##screenshot_button[vi] = Button(tab, text=name, bg="#d3d3d3", fg="black", command=on_screenshot, width=len(name)-width_adj-2)
  width = get_button_width(name, -3, 0, 0)

  screenshot_button[vi] = Button(tab, text=name, bg="white", fg="green", command=on_screenshot, highlightbackground="#e3e3e3", width=width)

  curr_x_pos = 10
  screenshot_button[vi].place(x=curr_x_pos, y=top_pos+(font_height+4)*2, height=font_height)

  old_name = name
  line_width = curr_x_pos + 10 + len(name)*font_width

  screenshot_combo[vi] = ttk.Combobox(tab, width=2)
  curr_x_pos = line_width - 10

  screenshot_combo[vi].place(x=curr_x_pos, y=top_pos+(font_height+4)*2)

  screenshot_combo[vi]['values'] = ("5", "10", "30")

  line_width = curr_x_pos + 10 + font_width*4

  ##screenshot_secs_label[vi] = Label(tab, text="secs")
  screenshot_secs_label[vi] = ttk.Label(tab, text="secs")
  curr_x_pos = line_width
  screenshot_secs_label[vi].place(x=curr_x_pos, y=top_pos+(font_height+4)*2)

  #----------------------------------------------------------------------
  # Show CD
  #----------------------------------------------------------------------
  name = "Show CD"
  show_cd_width = len(name)-2

  ##show_cd_button[vi] = Button(tab, text=name, bg="#d3d3d3", fg="black", command=on_show_cd, width=show_cd_width)
  width = get_button_width(name, -1, 8, 0)

  show_cd_button[vi] = Button(tab, text=name, bg="white", fg="black", command=on_show_cd, highlightbackground="#e3e3e3", width=width)

  curr_x_pos = start_vc_x_pos
  show_cd_button[vi].place(x=curr_x_pos, y=top_pos+(font_height+4)*2, height=font_height)

  line_width = curr_x_pos
  old_name = name

  #----------------------------------------------------------------------
  # Add CD
  #----------------------------------------------------------------------
  name = "Add CD"
  ##add_cd_button[vi] = Button(tab, text=name, bg="#d3d3d3", fg="black", command=on_add_cd, width=len(name)-2)
  width = get_button_width(name, 0, 8, 0)

  add_cd_button[vi] = Button(tab, text=name, bg="white", fg="black", command=on_add_cd, highlightbackground="#e3e3e3", width=width)

  curr_x_pos = line_width + 16 + (show_cd_width+2)*font_width
  add_cd_x_pos = curr_x_pos

  add_cd_button[vi].place(x=curr_x_pos, y=top_pos+(font_height+4)*2, height=font_height)

  line_width = curr_x_pos

  add_cd_combo[vi] = ttk.Combobox(tab)
  curr_x_pos = line_width + 16 + len(old_name)*font_width
  ##add_cd_combo[vi].place(x=curr_x_pos, y=top_pos+(font_height+5)*2, height=font_height)
  add_cd_combo[vi].place(x=curr_x_pos, y=top_pos+(font_height+5)*2)

  add_cd_combo[vi]['values'] = ("/usr/share/virtualbox/VBoxGuestAdditions.iso", "/classroom/iso/ubuntu-18.04.2-live-server-amd64.iso", "/classroom/iso/Fedora-Server-dvd-x86_64-30-1.2.iso")

  cd_combo_list = []

  cd_combo_list.append("")
  cd_combo_list.append("/usr/share/virtualbox/VBoxGuestAdditions.iso")
  cd_combo_list.append("/classroom/iso/ubuntu-18.04.2-live-server-amd64.iso")
  cd_combo_list.append("/classroom/iso/Fedora-Server-dvd-x86_64-30-1.2.iso")

  add_cd_combo[vi]['values'] = cd_combo_list
  add_cd_combo[vi].current(0)

  #----------------------------------------------------------------------
  # Eject CD
  #----------------------------------------------------------------------
  name = "Eject CD"
  ##eject_cddvd_button[vi] = Button(tab, text=name, bg="#d3d3d3", fg="black", command=on_eject_cddvd, width=len(name)-4)
  width = get_button_width(name, -2, 0, 0)

  eject_cddvd_button[vi] = Button(tab, text=name, bg="white", fg="black", command=on_eject_cddvd, highlightbackground="#e3e3e3", width=width)

  curr_x_pos = add_cd_x_pos
  eject_cddvd_button[vi].place(x=curr_x_pos, y=top_pos+(font_height+4)*3, height=font_height)

  #----------------------------------------------------------------------
  # Reveal password
  #----------------------------------------------------------------------
  name = "Reveal"
  ##reveal_pw_button[vi] = Button(tab, text=name, fg="black", command=on_reveal_pw, width=len(name)-3)
  width = get_button_width(name, -1, 4, 0)

  reveal_pw_button[vi] = Button(tab, text=name, bg="white", fg="black", command=on_reveal_pw, highlightbackground="#e3e3e3", width=width)

  curr_x_pos = 10
  reveal_pw_button[vi].place(x=curr_x_pos, y=top_pos+(font_height+4)*3, height=font_height)

  line_width = curr_x_pos
  old_name = name

  #----------------------------------------------------------------------
  # Paste password
  #----------------------------------------------------------------------
  name = "Paste"
  ##paste_pw_button[vi] = Button(tab, text=name, fg="black", command=on_paste_pw, width=len(name)-3)
  width = get_button_width(name, -1, 4, 0)

  paste_pw_button[vi] = Button(tab, text=name, bg="white", fg="black", command=on_paste_pw, highlightbackground="#e3e3e3", width=width)

  curr_x_pos = line_width + 10 + len(old_name)*font_width
  paste_pw_button[vi].place(x=curr_x_pos, y=top_pos+(font_height+4)*3, height=font_height)

  line_width = curr_x_pos
  old_name = name

  #----------------------------------------------------------------------
  # Guest User
  #----------------------------------------------------------------------
  guest_user_text[vi] = StringVar(tk_root, guest_user[vi] + " password")
  guest_user_label[vi] = ttk.Label(tab, textvariable=guest_user_text[vi])
  curr_x_pos = line_width + 10 + len(old_name)*font_width
  guest_user_label[vi].place(x=curr_x_pos, y=top_pos+(font_height+4)*3+2)

  line_width = curr_x_pos

  #----------------------------------------------------------------------
  # NAT Port Mapping
  #----------------------------------------------------------------------
  il_frame[vi] = ttk.Frame(tab)
  curr_x_pos = 10
  il_frame[vi].place(x=curr_x_pos, y=top_pos+(font_height+4)*4+2)

  info_list_height=6
  width = get_listbox_width(35, 10, 0, 0)

  info_list[vi] = Listbox(il_frame[vi], height=info_list_height, width=width)
  info_list[vi].grid(column=0, row=0, sticky='nsew')
###  info_list[vi].pack(side = "top", fill = "both", expand = True)

  il_scrollbar[vi] = ttk.Scrollbar(il_frame[vi], orient='horizontal', command=info_list[vi].xview)
  il_scrollbar[vi].grid(column=0, row=1, sticky='ew')
  ### il_scrollbar[vi].pack(side = "bottom", fill = "y")
  ### il_scrollbar[vi].place(x=curr_x_pos, y=top_pos+(font_height+10)*info_list_height)

  info_list[vi].configure(xscrollcommand=il_scrollbar[vi].set)

  #---------------------------------------------------------------------
  # Place Port Mapping info into list
  #---------------------------------------------------------------------
  port_map_y = 95
  port_info = ""
  ##tk_root.style.configure('TButton', background="#d3d3d3", foreground="black",
  ##                       anchor=N, padding=[0,0,0,0])
    
  for j in range(next_port_index[vi]+1):
    port_info = guest_service[vi][j]  + " "
    hostinfo = host_and_port[vi][j].split(":")
    
    #-------------------------------------------------------------------
    # If well-known services, then show guest and host information
    #-------------------------------------------------------------------
    if (guest_service[vi][j] == "ssh") or \
       (guest_service[vi][j] == "rdp") or \
       (guest_service[vi][j] == "ms-wbt-server"):
      port_info = port_info + " - guest NIC " + guest_nic[vi][j] + " port: " + guest_port[vi][j] + "; host port: " + hostinfo[1] + " -->"
      info_list[vi].insert(END, port_info)

      #-----------------------------------------------------------------
      # Some port-mapped services should provide a button to copy the
      # command with service-specific information as well as host and 
      # port information to the clipboard to avoid typos by users.
      #-----------------------------------------------------------------
      if guest_service[vi][j] == "ssh":
        name = "Copy ssh -p " + hostinfo[1] + " " + guest_user[vi] + "@" + hostinfo[0]
        copy_host_and_port_button[vi][j] = ttk.Button(tab, text=name, command=partial(on_copy_host_and_port, j), width=len(name)-width_adj)
        curr_x_pos = add_cd_x_pos
        copy_host_and_port_button[vi][j].place(x=curr_x_pos, y=port_map_y+18+font_height*j+3, height=font_height+2)

      elif (guest_service[vi][j] == "rdp") or \
           (guest_service[vi][j] == "ms-wbt-server"):
          name = "Copy " + host_and_port[vi][j]
          copy_host_and_port_button[vi][j] = ttk.Button(tab, text=name, command=partial(on_copy_host_and_port, j), width=len(name)-width_adj)
          curr_x_pos = add_cd_x_pos
          copy_host_and_port_button[vi][j].place(x=curr_x_pos, y=port_map_y+18+font_height*j+3, height=font_height+2)

      else:
        #---------------------------------------------------------------
        # If not well-known services, then create a generic button to 
        # copy host and port info.
        #---------------------------------------------------------------
        port_info = port_info + " - guest NIC " + guest_nic[vi][j] + " port: " + guest_port[vi][j] + "; host port: " + host_and_port[vi][j]
        info_list[vi].insert(END, port_info)

        name = "Copy " + host_and_port[vi][j]
        copy_host_and_port_button[vi][j] = ttk.Button(tab, text=name, command=partial(on_copy_host_and_port, j), width=len(name)-width_adj)
        curr_x_pos = add_cd_x_pos
#        copy_host_and_port_button[vi][j].place(x=curr_x_pos, y=port_map_y+18+font_height*j+3, height=font_height+2)
        copy_host_and_port_button[vi][j].place(x=curr_x_pos, y=port_map_y+18+font_height*j+3)

  if port_info == "":
    port_info = "(no pre-defined port mapping)"
    info_list[vi].insert(END, port_info)
  else:
    port_info = "NAT Port Mapping Information"
    info_list[vi].insert(0, port_info)

    #-------------------------------------------------------------------
    # Label to tell user what port-mapping "Copy" buttons will do.
    #-------------------------------------------------------------------
    ##port_map_label[vi] = Label(tab, text="Copy Relevant Port Mapping Info to Clipboard")
    port_map_label[vi] = ttk.Label(tab, text="Copy Relevant Port Mapping Info to Clipboard")
    curr_x_pos = add_cd_x_pos
    port_map_label[vi].place(x=curr_x_pos, y=port_map_y+1)

  #----------------------------------------------------------------------
  # Do cmd
  #----------------------------------------------------------------------
  name = "Do cmd:"
  ##do_cmd_button[vi] = Button(tab, text=name, bg="#d3d3d3", fg="black", command=on_do_cmd, width=len(name)-3)
  width = get_button_width(name, -1, 6, 0)

  do_cmd_button[vi] = Button(tab, text=name, bg="white", fg="black", command=on_do_cmd, highlightbackground="#e3e3e3", width=width)

  curr_x_pos = 10
  do_cmd_button[vi].place(x=curr_x_pos, y=top_pos+(font_height+4)*9+1, height=font_height)

  line_width = curr_x_pos
  old_name = name

  cmd_combo[vi] = ttk.Combobox(tab)
  curr_x_pos = line_width + 10 + len(old_name)*font_width + 5
  cmd_combo[vi].place(x=curr_x_pos, y=top_pos+(font_height+4)*9+1, height=font_height+5, width=50*font_width)

  line_width = curr_x_pos

  #----------------------------------------------------------------------
  # Set up the list of commands for the dropdown box to right of "Do cmd:"
  # 
  # The idea is to provide help for common commands that students may
  # need, categorized to make them a little easier to see.
  #----------------------------------------------------------------------
  start_vde = ""
  config_vde = "config_vde " + vm_name[vi]

  cmd_list = []

  cmd_list.append("")
  cmd_list.append("# ----- General commands -----")
  cmd_list.append("modifyvm " + vm_name[vi] + " ")
  cmd_list.append("controlvm " + vm_name[vi] + " ")
  cmd_list.append("# ----- Network-related commands -----")
  cmd_list.append("show_nic " + vm_name[vi])
  cmd_list.append("modifyvm " + vm_name[vi] + " --nic1 nat")
  cmd_list.append("modifyvm " + vm_name[vi] + " --nic1 intnet --intnet1 " + vcl_fn_user + "_intnet")
  cmd_list.append("modifyvm " + vm_name[vi] + " --nic1 hostonly --hostonlyadapter1 vboxnet0")
  cmd_list.append("# ----- Advanced Network-related commands -----")
  cmd_list.append(config_vde)
  cmd_list.append(start_vde)
  cmd_list.append("list dhcpservers")
  cmd_list.append("dhcpserver modify --netname " + vcl_fn_user + "_intnet --disable")
  cmd_list.append("dhcpserver modify --netname " + vcl_fn_user + "_intnet --enable")
  cmd_list.append("controlvm " + vm_name[vi] + " nicpromisc1 allow-all")
  cmd_list.append("controlvm " + vm_name[vi] + " nicpromisc1 deny")
  cmd_list.append("# ----- Media-related commands -----")
  cmd_list.append("add_cd " + vm_name[vi] + " ")
  cmd_list.append("# ----- Snapshot-related commands -----")
  cmd_list.append("snapshot " + vm_name[vi] + " list")
  cmd_list.append("snapshot " + vm_name[vi] + " take " + vcl_fn_user + "_changes")
  cmd_list.append("snapshot " + vm_name[vi] + " restore " + vcl_fn_user + "_changes")
  cmd_list.append("# ----- Revert to original VM -----")
  cmd_list.append("snapshot " + vm_name[vi] + " restore User_Changes")

  cmd_combo[vi]['values'] = cmd_list
  cmd_combo[vi].current(0)

  #----------------------------------------------------------------------
  # Set the title to use for this VM, which is the file name of the RDP file.
  # It appears as the RDP window title, and helps the user to see which
  # window is associated with what VM. Since there may be multiple
  # VMs with the same name, we add the index number to make sure
  # the user knows which one this connection deals with.
  #----------------------------------------------------------------------
  rdp_title[vi] = host_user[vi] + "_" + vm_name[vi] + "_" + str(vi) 
  rdp_fn = re.sub(r'\\', r'\\\\', os.path.join(get_profile_dir() , rdp_title[vi] + ".rdp"))

  #----------------------------------------------------------------------
  # If at the time of GUI creation an RDP file already exists for
  # the VM, it is deleted so that the latest information from the VCL
  # file is always used (vs. an older RDP file that may no longer have
  # the correct information in it).
  #----------------------------------------------------------------------
  if os.path.exists(rdp_fn) == True:
     os.remove(rdp_fn)

#------------------------------------------------------------------------------
# On close
#------------------------------------------------------------------------------
def on_close():
  exit_program()

#-------------------------------------------------------------------------
# When the Connect button is clicked, determine if the corresponding
# VM is running. If it is not running, issue a message stating that. 
# If the VM is running, connect to its display if possible, using whatever
# RDP method (guest or virtual) that the user desires to try.
#-------------------------------------------------------------------------
def on_connect():
  global rdp_info, entry_item

  vi = get_vi(connect_button)

  #----------------------------------------------------------------------
  # If the VM is NOT running
  #----------------------------------------------------------------------
  if is_running(vi) == 0:
    msgbox_info("The VM " + vm_name[vi] + " is not running, and hence you cannot connect to its display via remote" + get_crlf() + "desktop protocol (RDP). Start the VM." + get_crlf() + get_crlf() + "If no RDP method is available, ***after starting the VM***, try using the following" + get_crlf() + "commands via the 'Do cmd:' dropbox:" + get_crlf() + get_crlf() + "screenshot " + vm_name[vi] + " every 5" + get_crlf() + "sendkeys " + vm_name[vi] + " ... (specify keys and strings)")
  
  else: # VM is running
    #--------------------------------------------------------------------
    # Construct the absolute path to the RDP file.
    #--------------------------------------------------------------------
    rdp_fn = re.sub(r'\\', r'\\\\', os.path.join(get_profile_dir(), rdp_title[vi] + ".rdp"))

    # Indicate unknown rdp_info
    have_rdp_info = -1

    #--------------------------------------------------------------------
    # If user wants to change anything
    #--------------------------------------------------------------------
    if check_edit[vi].get() == 1:
      #--------------------------------------------------------------------
      # If Guest RDP information is desired
      #--------------------------------------------------------------------
      if check_rdp[vi].get() == 1:
        # Pick up default information
        have_rdp_info = setup_rdp_from_nic_info(vi)

        if have_rdp_info == 1: # If there was RDP information in NICs
          #---------------------------------------------------------------
          # Ask for changes to RDP info
          #---------------------------------------------------------------
          connect_input = extended_input_box("RDP Information (required=*)", \
                ["*RDP IP:", \
                 "*Port:", \
                 "Domain:", \
                 "*User (orig: '" + orig_guest_user[vi] + "'):", \
                 "*Password (orig: '" + orig_guest_pw[vi]  + "'):"], \
                [rdp_info[vi][ri_host], \
                 rdp_info[vi][ri_port], \
                 rdp_info[vi][ri_domain], \
                 guest_user[vi], \
                 guest_pw[vi]], \
                ["Update", "Okay"])

          #---------------------------------------------------------------
          # If user clicks update; write/rewrite info for the .rdp file
          #---------------------------------------------------------------
          if connect_input == True:
            rdp_info[vi][ri_host] = entry_item[0]
            rdp_info[vi][ri_port] = entry_item[1]
            rdp_info[vi][ri_domain] = entry_item[2]
            rdp_user[vi] = entry_item[3]
            rdp_pw[vi] = entry_item[4]
          
            #-------------------------------------------------------------
            # Update the guest user password to reveal/paste.
            #-------------------------------------------------------------
            guest_user_text[vi].set(guest_user[vi] + " password")
  
          else:
            rdp_user[vi] = guest_user[vi]
            rdp_pw[vi] = guest_pw[vi]

      #--------------------------------------------------------------------
      # If Connect Edit Checkbox is checked but VRDP is checked
      #--------------------------------------------------------------------
      elif check_rdp[vi].get() == 2: # If VRDP Info desired
        msgbox_info("Editing of VRDP information is not allowed.")
        return
  
    #--------------------------------------------------------------------
    # Otherwise, user does not want to change RDP info
    #--------------------------------------------------------------------
    else:
      if check_rdp[vi].get() == 2: # If VRDP Info desired
        if vrdp_addr[vi] != "":
          have_rdp_info = 1
          host_or_port = vrdp_addr[vi].split(":")

          rdp_info[vi][ri_host] = host_or_port[0]
          rdp_info[vi][ri_port] = host_or_port[1]
          rdp_info[vi][ri_domain] = ""
          rdp_user[vi] = host_user[vi]
          rdp_pw[vi] = host_pw[vi]
  
    #--------------------------------------------------------------------
    # If no RDP file for VM
    #--------------------------------------------------------------------
    if os.path.exists(rdp_fn) == False:
      #--------------------------------------------------------------------
      # If haven't tried to get latest RDP info
      #--------------------------------------------------------------------
      if check_rdp[vi].get() == 1: # If Guest RDP Info desired
        if have_rdp_info == -1:
          #----------------------------------------------------------------
          # Get the latest guest RDP info using NIC info, if any
          #----------------------------------------------------------------
          have_rdp_info = setup_rdp_from_nic_info(vi)
  
          rdp_user[vi] = guest_user[vi]
          rdp_pw[vi] = guest_pw[vi]

    else: # We have saved RDP info
      have_rdp_info = 2

    if have_rdp_info >= 1:
      if platform.system() == "Windows":
        #--------------------------------------------------------------------
        # Show progress of connecting; otherwise, nothing happens until the
        # connection window pops up, and the user may start to wonder what
        # is going on during those several seconds.
        #--------------------------------------------------------------------
        if platform.system() == "Windows":
          progress_on("Connecting...")

        pw_hash = hash_pw(rdp_pw[vi])

        if platform.system() == "Windows":
          progress_set(50.0)
    
        write_rdp_file(rdp_fn, rdp_info[vi][ri_host] + ":" + rdp_info[vi][ri_port], rdp_info[vi][ri_domain], rdp_user[vi], pw_hash)
        if platform.system() == "Windows":
          progress_set(75.0)

        pid = spawn_cmd('mstsc' + ' "' + rdp_fn + '"')

        if platform.system() == "Windows":
          progress_set(100.0)
          # Wait a little to allow the user to see the progress completion
          time.sleep(0.5)
  
          # Destroy the progress window
          progress_off()

      elif platform.system() == "Darwin":
        pc_name = rdp_user[vi] + "_" + vm_name[vi] + "_" + str(vi)

        tk_root.clipboard_clear()
        tk_root.clipboard_append(host_pw[vi])

        run_cmd("open >/dev/null 2>&1 --stdout /dev/null --stderr /dev/null -a \"Microsoft Remote Desktop\"", False, False)

        if os.path.exists(rdp_fn) == False:
          write_rdp_file(rdp_fn, rdp_info[vi][ri_host] + ":" + rdp_info[vi][ri_port], rdp_info[vi][ri_domain], rdp_user[vi], '')

##        msgbox_info("Use the following information in Microsoft Remote Desktop/Edit PC\n\n" + \
##                    "PC name: " + rdp_info[vi][ri_host] + ':' + rdp_info[vi][ri_port] + "\n" + \
##                    "User account: " + rdp_user[vi] + "\n" + \
##                    "Friendly name: " + vm_name[vi] + "\n" + \
##                    "\nWhen prompted for a password, use:\n\n" + rdp_pw[vi] + "\n\n" + \
##                    "Save it, then click on the " + vm_name[vi] + " icon to connect to the display.")

          msgbox_info("Using Microsoft Remote Desktop:\n\n" + \
                      "A. If nothing in your VM info changed and you have a PC named \"" +  pc_name + "\", double-click\n" + \
                      "     on that name to Connect.\n\n" +\
                      "    *** If the remote display shows a black window, you may also need to click inside the window and/or\n" + \
                      "         press the enter key to wake up the guest OS. ***\n\n" + \
                      "B. Otherwise, if there is no such PC name OR your VM info changed:\n\n" +\
                      "  1. If you need to update info in an existing PC \"" + pc_name +"\", delete the old one first:\n\n" + \
                      "    a. click on the PC icon\n" + \
                      "    b. click on the trash can icon for that PC.\n\n" + \
                      "  2. Select Connections menu item 'Import from RDP File...'\n\n" + \
                      "  3. Browse to the .rdp file:\n\n" + \
                      "       " +  rdp_fn + "\n\n" + \
                      "      and select it.\n\n" + \
                      "  4. Double-click on the PC Name \"" + pc_name + "\" icon to connect to the display.\n\n" + \
                      "  5. If it won't connect: \n\n" + \
                      "    a. edit that PC name (click on PC icon then pencil icon)\n" + \
                      "    b. change the 'User account:' settings to 'Ask when required'\n" + \
                      "    c. click 'Save'\n" + \
                      "    d. double-click on the PC name to connect.\n" + \
                      "    e. when prompted, provide 'Username:' value: " + rdp_user[vi] + "\n" + \
                      "    f. in the 'Password:' field, paste the password:\n\n" + \
                      "        " + host_pw[vi] + "\n\n" + \
                      "       if it still is on the clipboard\n" + \
                      "    g. click 'Continue' to connect." )
        else:
          msgbox_info("In Microsoft Remote Desktop app, double-click on PC named \"" + pc_name + "\"\n")

      elif platform.system() == "Linux":
        # Note: may be able to use a standard mstsc-like RDP file, perhaps
        # without the saved password
        cmd = "xfreerdp" + \
             " /v:" + rdp_info[vi][ri_host] + ':' + rdp_info[vi][ri_port] + \
             " /u:" + rdp_user[vi] + \
             " /p:'" + rdp_pw[vi] + "'" + \
             " /w:" + vcl_desktop_width + \
             " /h:" + vcl_desktop_height + \
             " /clipboard"

        tk_root.clipboard_clear()
        tk_root.clipboard_append(cmd)

        msgbox_info("This command is on the clipboard:\n\n" + cmd)

      else: # Not windows or macos
        msgbox_error("No OS Connect support for " + platform.system())
  
    else: # Not enough or any RDP information to continue, so tell user
      msgbox_info("The VM " + vm_name[vi] + " is not set up for remote RDP connections (i.e., no VRDP, or all enabled NICs do not have either a NAT or bridged network method with a cable connected, or the guest OS is not fully up yet)." + get_crlf() + get_crlf() +  "Use this command in the 'Do cmd:' dropdown box: " + "show_nic " +  vm_name[vi] + get_crlf() + get_crlf() + "to show current NIC information.")
  
#---------------------------------------------------------------------------
# When the Edit checkbox is checked
#---------------------------------------------------------------------------
def on_edit():
  vi = get_vi(connect_grdp)

  if check_rdp[vi].get() == 2: # If VRDP
    msgbox_info("Editing of VRDP information is not allowed.")
    check_edit[vi].set(0) # Turn off Edit
  
#---------------------------------------------------------------------------
# When the an RDP radio button is clicked, if an RDP file already
# exists, then delete it so that the latest guest RDP info can be used.
#---------------------------------------------------------------------------
def on_rdp():
  have_problem = False

  vi = get_vi(connect_grdp)
  
  if check_rdp[vi].get() == 2: # If VRDP
    if vrdp_addr[vi] == "":
      msgbox_info("VRDP information is not available for this VM.")
      have_problem = True

    else:
      check_edit[vi].set(0) # Turn off edit since VRDP was chosen

  if have_problem == True:
    check_rdp[vi].set(1) # Set to Guest RDP
    return

  rdp_fn = re.sub(r'\\', r'\\\\', os.path.join(get_profile_dir(), rdp_title[vi] + ".rdp"))

  if os.path.exists(rdp_fn) == True:
    os.remove(rdp_fn)
  
#-------------------------------------------------------------------------
# When the Start button is clicked, send a command to start the VM without
# any display, since it is remote and either VRDP, guest RDP or ssh
# will be used to interact.
#-------------------------------------------------------------------------
def on_start_vc():
  vi = get_vi(start_vc_button)

  send_vcl_command(vi, auth + "start " + vm_name[vi], 0, True)

#-------------------------------------------------------------------------
# When the Show State button is clicked, send a command to the host
# to show the VM's state.
#-------------------------------------------------------------------------
def on_show_state_vc():
  vi = get_vi(show_state_vc_button)

  send_vcl_command(vi, "vcl_status " + vm_name[vi], 0, True)

#-------------------------------------------------------------------------
# When the Set NIC button is clicked, send a command to the host to
# set the NIC of the running VM as defined in the list item chosen.
#-------------------------------------------------------------------------
def on_set_nic():
  vi = get_vi(set_nic_button)

  send_vcl_command(vi, auth + "controlvm " + vm_name[vi] + " " + nic_value[vi], 0, False)
  
#-------------------------------------------------------------------------
# When the Set NIC combination control is used, read what item (NIC
# setting) was selected and associate it with this VM.
#-------------------------------------------------------------------------
def on_nic_combo():
  vi = get_vi(nic_combo)

  nic_value[vi] =  nic_combo[vi].get()
  
#-------------------------------------------------------------------------
# When the C-A-End button is clicked, send Ctrl-Alt-End to the VM by using
# the send_cad command.
#-------------------------------------------------------------------------
def on_send_cae():
  vi = get_vi(cae_button)
  
  send_vcl_command(vi, auth + "send_cad " + vm_name[vi], 0, False)
  
#-------------------------------------------------------------------------
# When the Power Off button is clicked, first confirm with the user
# that that is what is desired, and if so, send the poweroff command to
# the host for this VM.
#-------------------------------------------------------------------------
def on_power_off():
  vi = get_vi(power_off_button)

  confirm_power_off = msgbox_yesno("Do you really want to power off '" + vm_name[vi] + "'?")

  if msgbox_yes == False:
    return
  else:
    send_vcl_command(vi, auth + "controlvm " + vm_name[vi] + " poweroff", 0, True)
  
#-------------------------------------------------------------------------
# If the Screenshot combination control is used, read the associated
# number of seconds between screenshots for later use.
#-------------------------------------------------------------------------
def on_screenshot_combo():
  # This entire function might be useless

  vi = get_vi(screenshot_combo)

  screenshot_secs = screenshot_combo[vi].get()
  
#-------------------------------------------------------------------------
# When the Screenshot button is clicked, read the current choice for
# the number of seconds between screenshots. If a valid number of seconds,
# use it to set up and control the screenshotting.
#-------------------------------------------------------------------------
def on_screenshot():
  vi = get_vi(screenshot_button)

  screenshot_secs = str(screenshot_combo[vi].get())

  # If none, default to 5
  if screenshot_secs == "":
    msgbox_error("Please choose number of seconds from dropdown box.")
    return

  #-----------------------------------------------------------------------
  # If a whole number
  #-----------------------------------------------------------------------
  if screenshot_secs.isdigit() == True:
    #---------------------------------------------------------------------
    # Stop any current periodic screenshots
    #---------------------------------------------------------------------
    lines = send_vcl_command(vi, "screenshot " + vm_name[vi] +  " stop ", 1, False)

    #---------------------------------------------------------------------
    # Execute the new screenshot command, returning the URL in output
    #---------------------------------------------------------------------
    lines = send_vcl_command(vi, "screenshot " + vm_name[vi] + " every " + screenshot_secs, 1, False)

    line = ""

    for oline in lines.splitlines():
      line = oline.decode('utf-8')
      line.rstrip() # Remove trailing newlines and whitespace
      break # First line only

    if lines == "": # If no lines
      msgbox_info("No information returned from screenshot.")

    else:
      value = line.split("://")
      
      #-------------------------------------------------------------------
      # If delimiter found, there will be two strings
      #-------------------------------------------------------------------
      if len(value) == 2:
        url = "http://" + value[1]

        #-----------------------------------------------------------------
        # Put the URL on the clipboard for easier pasting in web browser.
        #-----------------------------------------------------------------
        tk_root.clipboard_clear()
        tk_root.clipboard_append(url)
        msgbox_info("On clipboard: " + url)

      else:
        #-----------------------------------------------------------------
        # Otherwise, let's show them the error output
        #-----------------------------------------------------------------
        show_any_error("Error running screenshot command:", "", lines, True)

  else:
    msgbox_error("Invalid number of seconds: " + screenshot_secs)
  
#-------------------------------------------------------------------------
# When the Eject CD button is clicked, send a command to the host to
# eject the default attached CD.
#-------------------------------------------------------------------------
def on_eject_cddvd():
  vi = get_vi(eject_cddvd_button)

  send_vcl_command(vi, auth + "eject_cd " + vm_name[vi], 0, True)

#-------------------------------------------------------------------------
# When the Add CD button is clicked, send command to add CD for the
# value from the Add CD combination control.
#-------------------------------------------------------------------------
def on_add_cd():
  vi = get_vi(add_cd_button)

  send_vcl_command(vi, auth + "add_cd " + vm_name[vi] + " '" + add_cd_combo[vi].get() + "'", 0, True)
   
#-------------------------------------------------------------------------
# When the ADD CD Combination control is used, associate the value selected
# with the VM.
#-------------------------------------------------------------------------
def on_add_cd_combo():
  vi = get_vi(add_cd_combo)

  add_cd_value[vi] =  str(add_cd_combo[vi].get())

#-------------------------------------------------------------------------
# When the Show CD button is clicked, send a command to the host to show
# the attached CDs for the VM.
#-------------------------------------------------------------------------
def on_show_cd():
  vi = get_vi(show_cd_button)
  
  send_vcl_command(vi, auth + "show_cd " + vm_name[vi], 0, True)
  
#-------------------------------------------------------------------------
# When the Reveal PW button is clicked, put the password on the clipboard
# and show it to the user.
#-------------------------------------------------------------------------
def on_reveal_pw():
  vi = get_vi(reveal_pw_button)
 
  tk_root.clipboard_clear()
  tk_root.clipboard_append(guest_pw[vi])
  msgbox_info("On clipboard: " + guest_pw[vi])

#-------------------------------------------------------------------------
# When the Paste PW button is clicked, paste the password to the VM display's
# window.
#-------------------------------------------------------------------------
def on_paste_pw():
  vi = get_vi(paste_pw_button)
 
  #----------------------------------------------------------------------
  # Activate the VM display's window.
  #----------------------------------------------------------------------
##  activate_window(rdp_title[vi] + " - " + host_addr[vi])

##  time.sleep(0.2)

  #----------------------------------------------------------------------
  # Paste the password to the window.
  #----------------------------------------------------------------------
  keys = sanitize_keys(guest_pw[vi]) 
  send_keys(vi, "'\\''" + keys[1:len(keys)-1] + "'\\''%enter")
  
#-------------------------------------------------------------------------
# When the NAT Port Mapping Copy Host and Port button is present and
# clicked, put that information on the clipboard.
#-------------------------------------------------------------------------
def on_copy_host_and_port(pi):
  vi = get_vi(copy_host_and_port_button)

  # Put part of button label after "Copy " into clipboard
  string = copy_host_and_port_button[vi][pi].cget('text')
  tk_root.clipboard_clear()
  tk_root.clipboard_append(string[5:])
  
#-------------------------------------------------------------------------
# When the Do cmd Combination control is used, read the command from the
# control.
#-------------------------------------------------------------------------
def on_cmd_combo():
  # This entire function might be useless

  vi = get_vi(cmd_combo)

  last_cmd = cmd_combo[vi].get()
  
#-------------------------------------------------------------------------
# When the Do cmd button is clicked, get the command to run, prompt
# the user to determine if it is okay to run, and run it.
#-------------------------------------------------------------------------
def on_do_cmd():
  vi = get_vi(do_cmd_button)

  last_cmd = cmd_combo[vi].get()
  cmd_args = last_cmd.split(" ")

  #------------------------------------------------------------------------
  # If a paste command
  #------------------------------------------------------------------------
  if cmd_args[0] == "paste":
    #----------------------------------------------------------------------
    # Find or add the command to the command list for future reuse.
    #----------------------------------------------------------------------
    find_or_add_cmd_list(vi, last_cmd)

    #----------------------------------------------------------------------
    # Activate the VM display to paste into.
    #----------------------------------------------------------------------
    activate_window(rdp_title[vi] + " - " + host_addr[vi])

    time.sleep(0.2)

    #----------------------------------------------------------------------
    # Paste the string into the window.
    #----------------------------------------------------------------------
    string = last_cmd[len("paste")+1:]
    send_keys(vi, string)

    time.sleep(0.1)
  
  else:
    #----------------------------------------------------------------------
    # Replace problematic characters in command
    #----------------------------------------------------------------------
    last_cmd = re.sub("[\\\#\`\+\|]", "", last_cmd, 0)

    #----------------------------------------------------------------------
    # Find or add sanitized command to Do cmd Combination control for 
    # later reuse.
    #----------------------------------------------------------------------
    find_or_add_cmd_list(vi, last_cmd)

    cmd = last_cmd.split(" ")

    #----------------------------------------------------------------------
    # If user wants to use sendkeys command (ssh_vbm subcommand)
    #----------------------------------------------------------------------
    if cmd[0].lower() == "sendkeys":
      #--------------------------------------------------------------------
      # Show progress of sending keystrokes and executing them.
      #--------------------------------------------------------------------

      if len(cmd) > 2:
        if cmd[2].lower() != "help":
          if platform.system() == "Windows":
            progress_on("Sending keys...")

      #--------------------------------------------------------------------
      # Send the keys to the VM, immediately and without prompting for 
      # user confirmation.
      #--------------------------------------------------------------------
      lines = send_vcl_command(vi, auth + last_cmd, 1, True)

      if len(cmd) > 2:
        if cmd[2].lower() != "help":
          if platform.system() == "Windows":
            progress_set(100.0)
            time.sleep(0.5) # Allow user to see progress completion
            progress_off()

    else:
      #--------------------------------------------------------------------
      # Make sure the user didn't inadvertently click on Do cmd button to
      # execute the current command in the Do cmd Combination control.
      #--------------------------------------------------------------------
      msgbox_yesno("Do you really want to run this VBoxManage command?" + get_crlf()  + get_crlf()  +  "  " + last_cmd)

      #--------------------------------------------------------------------
      # If quit/no
      #--------------------------------------------------------------------
      if msgbox_no == True:
        return

      else:
        send_vcl_command(vi, auth + last_cmd, 0, True)

#-----------------------------------------------------------------------------
# Get directory of profile (home directory):
#-----------------------------------------------------------------------------
def get_profile_dir():
  if platform.system() == "Windows":
    return os.environ.get("USERPROFILE")
  return os.path.expanduser("~")

#-----------------------------------------------------------------------------
# Global variables and main program  
#-----------------------------------------------------------------------------
global mv
global mve
global mv_version
global mvv

mv = "Manage VC"
mve = mv + " Error"
mv_version = "4.1.10"
mvv = mv + " " + mv_version

global msgbox_ok
global msgbox_no
global msgbox_cancel

msgbox_ok = False
msgbox_no = False
msgbox_cancel = False

global main_window_height
global main_window_width
global main_window_adj

main_window_height = 290
if platform.system() == "Darwin":
  main_window_height = 310

main_window_width = 525
main_window_width_adj = 50

if platform.system() == "Darwin":
  main_window_width_adj = 330

main_window_width = main_window_width + main_window_width_adj

global ini_fn
global log_fn

ini_fn = "manage_vc.ini"
log_fn = "manage_vc.log"

global dt_width
global dt_height
global dt_bpp
global dt_mode

dt_width = 1024
dt_height = 768
dt_bpp = 24
dt_mode = 1

global ssh
global ssh_options
global plink_opt
global scp
global user
global filename
global server
global destination_dir
global vcl_fn_user
global vcl_file
global orig_user

ssh = "xxx"
ssh_options = ""
plink_opt = ""
scp = ""
user = ""
filename = ""
server = ""
destination_dir = ""
vcl_fn_user = ""
vcl_file = ""
orig_user = ""

global vcl_desktop_width
global vcl_desktop_height
global vcl_desktop_bpp
global vcl_desktop_mode

vcl_desktop_width = str(dt_width)
vcl_desktop_height = str(dt_height)
vcl_desktop_bpp = str(dt_bpp)
vcl_desktop_mode = str(dt_mode)

global auth
auth = ""

global vm_limit
global port_limit
global nic_limit
global nic_info_limit
global rdp_info_limit

vm_limit = 20
port_limit = 5
nic_limit = 8
nic_info_limit = 3
rdp_info_limit = 3

global ni_method
global ni_cc_status
global ni_rule

ni_method = 0
ni_cc_status = 1
ni_rule = 2

global ri_host
global ri_port
global ri_domain

ri_host = 0
ri_port = 1
ri_domain = 2

global vcl_fn

vcl_fn = ".vcl"

global cn_ip_info
cn_ip_info =  [None for _ in range(vm_limit)]

global host_user
global host_pw
global last_host_addr
global last_host_user
global last_host_pw

host_user =  [None for _ in range(vm_limit)]
host_pw =  [None for _ in range(vm_limit)]
last_host_addr=""
last_host_user = ""
last_host_pw = ""

global vm_name
global host_addr
global vrdp_addr
global orig_guest_user
global orig_guest_pw
global guest_user
global guest_pw
global rdp_info
global rdp_user
global rdp_pw

vm_name =  ["" for _ in range(vm_limit)]
host_addr =  [None for _ in range(vm_limit)]
vrdp_addr =  [None for _ in range(vm_limit)]
orig_guest_user =  [None for _ in range(vm_limit)]
orig_guest_pw = [None for _ in range(vm_limit)]
guest_user =  [None for _ in range(vm_limit)]
guest_pw =  [None for _ in range(vm_limit)]
rdp_info =  [[0] * rdp_info_limit for _ in range(vm_limit)]
rdp_user =  [None for _ in range(vm_limit)]
rdp_pw =  [None for _ in range(vm_limit)]

global host_and_port
global guest_port
global guest_nic
global guest_service
global nic_info
global next_port_index
global port_map_label

host_and_port =  [[0] * port_limit for _ in range(vm_limit)]
guest_port =  [[0] * port_limit for _ in range(vm_limit)]
guest_nic =  [[0] * port_limit for _ in range(vm_limit)]
guest_service =  [[0] * port_limit for _ in range(vm_limit)]

nic_info =  [[[0] * nic_info_limit for _ in range(nic_limit)] for _ in range(vm_limit)]

next_port_index =  [-1 for _ in range(vm_limit)]
port_map_label =  [None for _ in range(vm_limit)]

global vmi
vmi = 0

global mw
global host_combo
global host_combo_max_len

mw = ""
host_combo = ""
host_combo_max_len = 0

global tab_main
global tab
global tab_name
global tab_i

tab_main = ""
tab =  [None for _ in range(vm_limit)]
tab_name =  [None for _ in range(vm_limit)]
tab_i =  [None for _ in range(vm_limit)]

global connect_button
global connect_vrdp
global connect_grdp
global connect_edit
global screenshot_button
global screenshot_list
global screenshot_combo
global screenshot_secs_label
global show_state_vc_button
global set_nic_button
global start_vc_button
global cae_button
global power_off_button
global eject_cddvd_button
global add_cd_button
global show_cd_button
global reveal_pw_button
global paste_pw_button
global reveal_ip_button
global guest_user_label
global guest_user_text
global paste_vc_button
global check_ssh_button
global info_list
global il_frame
global il_scrollbar
global nic_combo
global nic_value
global add_cd_combo
global add_cd_value
global do_cmd_button
global cmd_combo
global rdp_title
global copy_host_and_port_button

connect_button = [None for _ in range(vm_limit)]
connect_vrdp =  [None for _ in range(vm_limit)]
connect_grdp =  [None for _ in range(vm_limit)]
connect_edit =  [None for _ in range(vm_limit)]
screenshot_button =  [None for _ in range(vm_limit)]
screenshot_list =  [None for _ in range(vm_limit)]
screenshot_combo =  [None for _ in range(vm_limit)]
screenshot_secs_label =  [None for _ in range(vm_limit)]
show_state_vc_button =  [None for _ in range(vm_limit)]
set_nic_button =  [None for _ in range(vm_limit)]
start_vc_button =  [None for _ in range(vm_limit)]
cae_button =  [None for _ in range(vm_limit)]
power_off_button =  [None for _ in range(vm_limit)]
eject_cddvd_button =  [None for _ in range(vm_limit)]
add_cd_button =  [None for _ in range(vm_limit)]
show_cd_button =  [None for _ in range(vm_limit)]
reveal_pw_button =  [None for _ in range(vm_limit)]
paste_pw_button =  [None for _ in range(vm_limit)]
reveal_ip_button =  [None for _ in range(vm_limit)]
guest_user_label =  [None for _ in range(vm_limit)]
guest_user_text =  [None for _ in range(vm_limit)]
paste_vc_button =  [None for _ in range(vm_limit)]
check_ssh_button =  [None for _ in range(vm_limit)]
info_list =  [None for _ in range(vm_limit)]
il_frame =  [None for _ in range(vm_limit)]
il_scrollbar =  [None for _ in range(vm_limit)]
nic_combo =  [None for _ in range(vm_limit)]
nic_value =  [None for _ in range(vm_limit)]
add_cd_combo =  [None for _ in range(vm_limit)]
add_cd_value =  [None for _ in range(vm_limit)]
do_cmd_button =  [None for _ in range(vm_limit)]
cmd_combo =  [None for _ in range(vm_limit)]
rdp_title =  [None for _ in range(vm_limit)]
copy_host_and_port_button =  [[0] * port_limit for _ in range(vm_limit)]

global check_rdp, check_edit
check_rdp = [None for _ in range(vm_limit)]
check_edit = [None for _ in range(vm_limit)]

global cmd_list
cmd_list = []

global ping_error
ping_error = ""

global font_width
global font_height
global crypt_path
global rdp_fn
global rdp_exec

if platform.system() == "Windows":
  font_width = 7
else:
  font_width = 8
font_height = 20

crypt_path = ""
rdp_fn = ""
rdp_exec = ""

global no_pause_output_file
no_pause_output_file = os.path.join(get_profile_dir(), "nopause.txt")

global log_file
log_file = ""

global vi
global name_and_addr
global name_and_addr_selected
global pw_hash

vi = 0
name_and_addr =  [None for _ in range(vm_limit)]
name_and_addr_selected = ""
pw_hash = ""

global entry_item
global entry_box

global pb_top, progress, progress_label
pb_top = None
progress = None

global key_map
key_map = {'%' : '%per', 
           '\\' : '%bs',
           "'" : '%sq',
           '`' : '%bt',
           ';' : '%sc',
           '&' : '%amp',
           '|' : '%or'
          }

global max_names_len
max_names_len=0

global cssgate_pw
cssgate_pw = None 

# End globals

#------------------------------------------------------------------------------
# messagebox functions
#------------------------------------------------------------------------------
def msgbox_info(msg):
  lines = msg.splitlines(True)
  if len(lines) < 5:
    messagebox.showinfo(mv + " Information", msg)
  else:
    scrollable_listbox(mv + " Information", msg.splitlines(True), False)

  set_msgbox_result(True)

def msgbox_error(msg):
  messagebox.showerror(mve, msg)
  set_msgbox_result(True)

def msgbox_yesno(msg):
  result = messagebox.askyesno(mv, msg)
  set_msgbox_result(result)

def msgbox_yesnocancel(msg):
  result = messagebox.askyesnocancel(mv, msg)
  set_msgbox_result(result)

#------------------------------------------------------------------------------
# Set up global variables for result of messagebox()
#------------------------------------------------------------------------------
def set_msgbox_result(result):
  global msgbox_yes, msgbox_no, msgbox_cancel, tk_root_state
  msgbox_yes = False
  msgbox_no = False
  msgbox_cancel = False

  if result is None:
    msgbox_cancel = True
  elif isinstance(result, bool):
    if result == True:
      msgbox_yes = True
    else:
      msgbox_no = True

  #----------------------------------------------------------------------------
  # Cause root window allowing messagebox() to exit its mainloop, to be resumed
  # later if all goes well prior to populating the GUI.
  #----------------------------------------------------------------------------
  tk_root.after(500,lambda:tk_root.quit()) 

  tk_root.mainloop() # Process window events
  tk_root_state = 1
  
#------------------------------------------------------------------------------
# input box functions
#------------------------------------------------------------------------------
def input_box(msg, init_value, show_value):
  global tk_root
  tk_root.focus_force()
  value = simpledialog.askstring(mvv, msg, parent=tk_root, initialvalue=init_value, show=show_value)
  return value

#---------------------------------------------------------------------------
# An extended input box ("eib") is one with multiple input boxes. There must
# be a one-to-one correspondence between the items in the text_list
# (the labels) and those of the default_list (the default values for
# the items). The button list represents valid choices, but only
# "Update" and "Okay" are supported -- this will return None if
# any other button is defined. 
#
# "Update" means that all of the boxes values will be extracted
# from their respective input boxes and placed in the entry_item list.
# True will be returned as the function's result.
#
# "Okay" means that nothing was changed (even if something was), and to
# accept whatever the defaults were. False will be returned as the function's
# result.
#
# A "close window" event causes the same behavior as "Okay".
#---------------------------------------------------------------------------
def extended_input_box(title, text_list, default_list, button_list):
  global eib_top, entry_box, eib_result, entry_item
  eib_result = None

  eib_top = Toplevel(tk_root)
  eib_top.title(title)
  eib_top.geometry(str(font_width*25*2)+'x'+str(len(default_list)*font_height+font_height*4))
  #----------------------------------------------------------------------------
  # In case user won't click on either button, they can close window and do 
  # nothing.
  #----------------------------------------------------------------------------
  eib_top.protocol("WM_DELETE_WINDOW", on_eib_close) 

  entry_box = []
  value = [None for _ in range(len(default_list))]

  row = 0
  for i in range(len(text_list)):
    label = Label(eib_top, text=text_list[i])
    label.grid(column=0, row=row)
    value[i] = StringVar(eib_top, default_list[i])
    ebox = Entry(eib_top, textvariable=value[i])
    ebox.grid(column=1, row=row)
    entry_box.append(ebox)
    row = row + 1
   
  column = 0 
  for i in range(len(button_list)):
    this_button = ttk.Button(eib_top, text=button_list[i], command=partial(on_eib_button, button_list[i])).grid(column=column, row=row)
    column = column + 1

  eib_top.mainloop() # Process EIB events
  return eib_result

#---------------------------------------------------------------------------
# When extended_input_box button is clicked
#---------------------------------------------------------------------------
def on_eib_button(name):
  global eib_top, eib_result, entry_box, entry_item

  if name == "Update":
    entry_item = []
    for i in range(len(entry_box)):
      entry_item.append(entry_box[i].get())
    eib_result = True

  elif name == "Okay":
    eib_result = False

  eib_top.destroy()
  eib_top.quit()

#---------------------------------------------------------------------------
# When extended_input_box window is closed
#---------------------------------------------------------------------------
def on_eib_close():
  global eib_top, eib_result
  if eib_result is None:
    eib_result = False
    eib_top.destroy()
    
#---------------------------------------------------------------------------
# Scrollable List Box
#---------------------------------------------------------------------------
def scrollable_listbox(title, info_list, decode):
  global slb_top, main_window_width, main_window_height

  slb_top = Toplevel(tk_root)
  slb_top.title(title)
  width = 60
  if platform.system() == "Windows":
    width = 140

  slb_top.geometry(str(main_window_width+width)+'x'+str(main_window_height))

  #----------------------------------------------------------------------------
  # In case user won't click on either button, they can close window and do 
  # nothing.
  #----------------------------------------------------------------------------
  slb_top.protocol("WM_DELETE_WINDOW", on_slb_close2) 

  width = get_listbox_width(80, 0, 0, 0)
  lb = Listbox(slb_top, height=15, width=width)
  lb.grid(column=0, row=0, sticky='nwes')

  sb = ttk.Scrollbar(slb_top, orient='vertical', command=lb.yview)
  sb.grid(column=1, row=0, sticky='ns')

  slbutton = ttk.Button(slb_top, text="Okay", command=lambda: on_slb_close(slb_top))
  slbutton.grid(column=0, row=1, sticky='s')
  slbutton.bind('<Return>', on_slb_close)

  lb.configure(yscrollcommand=sb.set)

  for item in info_list:
    if decode == True:
      lb.insert(END, ' ' + item.decode('utf-8'))
    else:
      lb.insert(END, ' ' + item)

  slbutton.focus_force()
  slb_top.mainloop()

#---------------------------------------------------------------------------
# On scrollable list box close
#---------------------------------------------------------------------------
def on_slb_close(self):
  global slb_top
  slb_top.destroy()

#---------------------------------------------------------------------------
# On scrollable list box close
#---------------------------------------------------------------------------
def on_slb_close2():
  global slb_top
  slb_top.destroy()

#---------------------------------------------------------------------------
# Find or add the command to the command list of the Do cmd Combination
# control. 
#---------------------------------------------------------------------------
def find_or_add_cmd_list(vi, last_cmd):
  global cmd_list, cmd_combo

  cmd_list = list(cmd_combo[vi]['values'])

  #-------------------------------------------------------------------------
  # Determine if last command is already in command list
  #-------------------------------------------------------------------------
  ci = -1
  for i in range(len(cmd_list)):
    if cmd_list[i] == last_cmd:
      ci = i
      break

  #-------------------------------------------------------------------------
  # If not in command list, add it
  #-------------------------------------------------------------------------
  if ci == -1:
    cmd_list.append(last_cmd)
    cmd_combo[vi]['values'] = cmd_list

#---------------------------------------------------------------------------
# Set up plink to send VCL commands, since special options may be required
# to do so without user confirmation of command.
#---------------------------------------------------------------------------
def setup_send_vcl_command():
  global plink_opt, ssh_options, scp, ssh

  plink_opt = ""
  if ssh.find("plink") > 0:
    if ssh_options == "":
      ssh_options = "-ssh -2"
  
    plink_opt = "-no-antispoof -pw "

  elif ssh.find("ssh") > 0:
    ssh_options = ssh_options + " -o StrictHostKeyChecking=no"
  
#---------------------------------------------------------------------------
# Capture output from ssh command
#---------------------------------------------------------------------------
def capture_output_from(cmd, echo_no, cmd_part, host_password):
  lines = ""

  if platform.system() == "Darwin":
    pw_file = os.path.join(get_profile_dir(), ".manage_vc.pw")

    pw_fh = open(pw_file, "w")
    pw_fh.write(host_password)
    pw_fh.close()

    cmd = "sshpass -f '" + pw_file + "' " + cmd

  cmd_args = shlex.split(cmd)

  not_yet_complete = False

  while not_yet_complete == False:
    process = popen_spawn.PopenSpawn(cmd_args)

    next_index = -1

    #--------------------------------------------------------------------
    # If this message, the host key is not saved for pscp or cygwin scp.
    #--------------------------------------------------------------------
    index = process.expect(['connection.', 'continue connecting', pexpect.TIMEOUT, pexpect.EOF], timeout=2)
#      print(f"cmd={cmd}\necho_no={echo_no}\nindex={index}\nnext_index={next_index}\n")

    if index == 0:
      if next_index != -1:
        msgbox_error("Problem capturing output from ssh command.")
        not_yet_complete = True

      process.sendline("y")
      process.sendeof()
      process.kill(signal.CTRL_C_EVENT)
      next_index = 3

    elif index == 1:
      process.sendline("yes")
      process.sendeof()
      process.kill(signal.CTRL_C_EVENT)

    elif index == 3:
      lines = process.before
          
      process.wait()
      not_yet_complete = True

    elif index == 2:
      process.sendline(host_password)
      process.sendeof()
      lines = process.before
          
      process.wait()
      not_yet_complete = True

  return lines

#---------------------------------------------------------------------------
# Show errors if there is something to show.
#---------------------------------------------------------------------------
def show_any_error(error, vcl_cmd, lines, show_output):
  error_msg = ""
  olines = ""

  if show_output == True:
    for oline in lines.splitlines():
       line = oline.decode('utf-8')
       olines = olines + line
       error_msg = error_msg + line + "\n"
    
    if olines != "":
      msgbox_error(error + "\n" + vcl_cmd + "\n" + error_msg)

#---------------------------------------------------------------------------
# Send a command to the VM's host to manage, query or set the VM.
#---------------------------------------------------------------------------
def send_vcl_command(vi, vcl_cmd, no_pause_option, show_output):
  global plink_option, vcl_cmd_title, tk_root, host_pw, ssh_options

  lines = ""

  echo_no = "echo n | " # default is to answer "no" to host key not found question of plink

  cn_ip_info = host_addr[vi].split(":")
  vcl_cmd_title = vcl_cmd
  cmd_part = vcl_cmd.split(" ")

  if plink_opt.find("-pw") > 0: # if password option, add password
    plink_option = plink_opt + host_pw[vi]
  else:
    plink_option = ""

  if ssh.find("plink") == -1:
    if ssh_options.find("StrictHostKeyChecking") < 0:
      ssh_options = ssh_options + " -o StrictHostKeyChecking=no"

  if have_host_key(cn_ip_info[0]) == True: # If host key found, don't echo anything
    echo_no = ""

  #---------------------------------------------------------------------------
  # Create the ssh command and the VCL command to manage the VM. Redirection
  # indicates if and how any output is to be captured.
  #---------------------------------------------------------------------------
  this_platform = ""

  if platform.system() == "Windows":
    this_platform = " __win" + mv_version + " "
  elif platform.system() == "Darwin":
    this_platform = " __mac" + mv_version + " "

  cmd = '"' + re.sub(r'\\', r'\\\\', ssh) + '" ' + ssh_options + ' ' + plink_option + ' ' + host_user[vi] + "@" + cn_ip_info[0] + ' ' + '@' + user + this_platform + vcl_cmd

  write_log(vi, cmd)

  #---------------------------------------------------------------------------
  # If no pause is desired, ensure the command executes and returns 
  # immediately afterwards.
  #---------------------------------------------------------------------------
  if no_pause_option == 1:
    lines = capture_output_from(cmd, echo_no, cmd_part, host_pw[vi])

    if lines != "":
      if (cmd_part[0] == "sendkeys") and (len(cmd_part) == 3) and (cmd_part[2] == "help"):
#        os_cmd = "cat"
#        if platform.system() == "Windows":
#          os_cmd = "type"

#        run_cmd(get_command_spec(' /d /c ') + quote(echo_no + os_cmd) + ' "' + no_pause_output_file + '"', True, False) 
        scrollable_listbox(vcl_cmd_title, lines.splitlines(), True)
        return lines 

      else:
        show_any_error("Error in:", vcl_cmd, lines, show_output)

  else:
    lines = capture_output_from(cmd, echo_no, cmd_part, host_pw[vi])
    if show_output == True:
      if (cmd_part[0] == "screenshot") and (len(cmd_part) == 3) and (cmd_part[2] == "showurl"):
        for oline in lines.splitlines():
          url = oline.decode('utf-8')
          if url.find("http") >= 0:
            tk_root.clipboard_clear()
            tk_root.clipboard_append(url)
            msgbox_info("On clipboard: " + url)
            lines = ""
            return lines

      scrollable_listbox(vcl_cmd_title, lines.splitlines(), True)

  return lines 

    #run_cmd(get_command_spec(' /d /c ') + quote(echo_no + cmd), True, False)
#    full_cmd = get_command_spec(' /d /c ') + quote(cmd) + pause_value 
#    cmd_args = shlex.split(full_cmd)
#    subprocess.run(cmd_args)

#---------------------------------------------------------------------------
# Run the command
#---------------------------------------------------------------------------
def run_cmd(cmd,  pause, pipe_stdout):
  global vcl_cmd_title
  ### print(f'***{cmd}***')
  try:
    if pause == True:
      output = subprocess.check_output(cmd, shell=True)
    else:
      retcode = subprocess.call(cmd, shell=True)

  except OSError as e:
    print("Execution failed:", e, file=sys.stderr)
  
  if pause == True:
    scrollable_listbox(vcl_cmd_title, output.splitlines(), True)

#---------------------------------------------------------------------------
# Spawn the command in new process
#---------------------------------------------------------------------------
def spawn_cmd(cmd):
###  print(f'***{cmd}***')
  try:
    process = subprocess.Popen(shlex.split(cmd))
  except subprocess.CalledProcessError as e:
    pass
  
  return process

#---------------------------------------------------------------------------
# Sanitize keys
#---------------------------------------------------------------------------
def sanitize_keys(orig_keys):
  global key_map

  keys = orig_keys

  if re.search('[%\\\'`;&|]', keys) is not None:
    new_keys = ""
    in_quote = False
    literal = "'\''"

    for i in range(len(keys)):
      try:
        k = key_map[keys[i:i+1]]
      except:
        k = None

      if k is not None: # If substitution occurred
        if in_quote == True:
          new_keys = new_keys + literal + "'\''"
          in_quote = False
          literal = "'\''"

        new_keys = new_keys + k
      else:
        if in_quote == False:
          in_quote = True
        literal = literal  + keys[i:i+1]

    if in_quote == True:
      new_keys = new_keys + literal + "'\''"
      in_quote = False

    keys = new_keys
  else:
    keys = "'" + keys + "'"

  return keys

#---------------------------------------------------------------------------
# Send keys
#---------------------------------------------------------------------------
def send_keys(vi, keys):
  lines = send_vcl_command(vi, auth + "'sendkeys " + vm_name[vi] + " " + keys +"'", 1, True)

  show_any_error("Error in:", "sendkeys " + vm_name[vi] + " " + keys, lines, True)

#---------------------------------------------------------------------------
# Write to log file using datetime stamp, user, VM name, and message
#---------------------------------------------------------------------------
def write_log(i, msg):
  global log_file

  if i == -1:
    vm_name = ""
  else:
    vm_name = tab_name[i]

  now = datetime.now()
  log_file.write(now.strftime("%Y%m%d_%H:%M:%S") + ' user="' + vcl_fn_user + '" vm="' + vm_name + '" ' + msg + '\n')
  log_file.flush()

#---------------------------------------------------------------------------
# Read lines from VCL file, parse the information, and store it in
# host-specific array and arrays of various VM-related information 
# indexed by the VM index.
#---------------------------------------------------------------------------
def read_vcl_file():
  global vmi, host_addr, last_host_addr, last_host_user, last_host_pw

  try:
    file = open(vcl_file, 'r')

  except OSError:
    msgbox_error("Unable to open file '" + vcl_file + "'")
    exit_program()

  while True:
    line = file.readline()
    if not line: # If EOF
      break

    line.rstrip() # Remove trailing newlines and whitespace

    #-------------------------------------------------------------------------
    # Parse line by blank-delimited values.
    #-------------------------------------------------------------------------
    value = line.split(" ")

    #-------------------------------------------------------------------------
    # If the third token contains "vcl"
    #-------------------------------------------------------------------------
    if value[2][:3] == "vcl":
      #-----------------------------------------------------------------------
      # If the fourth token contains "user", there is user information
      #-----------------------------------------------------------------------
      if value[3] == "user":
        last_host_user = value[4]
        last_host_pw = value[6].replace("'", "")[0:-1]
   
      #-----------------------------------------------------------------------
      # Else If the fourth token contains "host", there is multiple-valued
      # host information requiring extraction.
      #-----------------------------------------------------------------------
      elif value[3] == "host":
        extract_multi(vmi, value, -1)
   
      #-----------------------------------------------------------------------
      # Else presume there is multiple-valued information requiring extraction.
      #-----------------------------------------------------------------------
      else:
        extract_multi(vmi, value, -1)

    #-------------------------------------------------------------------------
    # Else If the third token is an or bar "|", that indicates a nested level of
    # information to extract.
    #-------------------------------------------------------------------------
    elif value[2] == "|":
      extract_multi(vmi, value, 0)

    #-------------------------------------------------------------------------
    # Else If the third token is a series of hyphens, and the fourth token
    # is "Information" , this is information for a new VM named in this line.
    #-------------------------------------------------------------------------
    elif value[2] == "---------":
      if value[3] == "Information": # If found information for VM
        vmi = find_vm_name_index(value[5])
        if last_host_addr == "":
          msgbox_error("Host IP address is not in the .vcl file")
          exit_program()

        host_addr[vmi] = last_host_addr # Associate host address with VM

  file.close()

#---------------------------------------------------------------------------
# Using the blank-delimited value array and an index offset, pick up the
# value at that index and offset and match it to what we are checking for.
# If found and VM-specific info, record that information for the VM whose 
# index is provided. Otherwise, put it where it belongs (e.g., host address).
#---------------------------------------------------------------------------
def extract_multi(vmi, value, offset):
  global vrdp_addr, orig_guest_user, guest_user, orig_guest_pw, guest_pw, host_user, host_pw
  global host_and_port, guest_service, guest_nic, guest_port
  global last_host_addr

  #-------------------------------------------------------------------------
  # If the value looks like the beginning of a host IP address
  #-------------------------------------------------------------------------
  if value[6+offset][:1] == "'":
    #-------------------------------------------------------------------------
    # The file can have multiple host IP addresses if user has more than one
    # course and .vcl information is appended. Pick up the latest one
    # and use from now on.
    #-------------------------------------------------------------------------
    last_host_addr = value[6+offset].replace("'", "")

  #-------------------------------------------------------------------------
  # Else If the value is related to VRDP info, extract the VRDP IP address
  # which usually is the same as the host IP address.
  #-------------------------------------------------------------------------
  elif value[6+offset] == "(vrdp)":
    vrdp_addr[vmi] = value[7+offset].replace("'", "")

  #-------------------------------------------------------------------------
  # Else If the value is related to guest login info, extract the user account
  # name and password (inside quotes). The host user and its password is 
  # the same for all VMs on the same host, so we associate them with the VM;
  # i.e., store them in the appropriate array indexed by the VM index.
  #-------------------------------------------------------------------------
  elif value[6+offset] == "(login)":
    orig_guest_user[vmi]= value[7+offset]
    guest_user[vmi]= orig_guest_user[vmi]
    orig_guest_pw[vmi]= value[9+offset].replace("'", "")[0:-1]
    guest_pw[vmi]= orig_guest_pw[vmi]
   
    host_user[vmi] = last_host_user
    host_pw[vmi] = last_host_pw

  #-------------------------------------------------------------------------
  # Else If the value is related to host port info (for NAT port mapping), 
  # save the port mapping info for later insertion into the list and into
  # a button that will copy host information to the clipboard for possible
  # pasting by the user.
  #-------------------------------------------------------------------------
  elif value[6+offset] == "host":
    next_pi = get_next_pi()
    host_and_port[vmi][next_pi] = value[8+offset].replace("'", "")
    guest_service[vmi][next_pi] = value[9+offset][1:len(value[9+offset])-1]
    guest_nic[vmi][next_pi] = value[13+offset]
    guest_port[vmi][next_pi] = value[15+offset][0:-1]

#---------------------------------------------------------------------------
# Get the next port index for this VM, checking to make sure it doesn't 
# exceed its array bounds.
#---------------------------------------------------------------------------
def get_next_pi():
  global next_port_index
  next_port_index[vmi] = next_port_index[vmi] + 1

  if next_port_index[vmi] == port_limit:
    msgbox_error("Exceeded maximum number of ports for VM '" + vm_name[vmi] + "': " + port_limit)
    exit_program()

  return next_port_index[vmi]

#---------------------------------------------------------------------------
# This function is called when a VM name is to be added, but there might
# be other VMs named the same. If there is another VM of the same name, its
# tab name is changed to VM_name/tabindex, and the search continues for
# all other duplicates. Once there are no more names to check, we add
# this name to vm_name and add its tab_name with its own unique tabindex.
#
# So, the first win10 VM will be saved as win10 in vm_name and have win10
# as a tab name, but when another win10 VM appears, the first win10's tab name
# will be changed to win10/1 and the second one's tab name will be win10/2.
#---------------------------------------------------------------------------
def find_vm_name_index(name):
  global vm_name, tab_name, tab_i

  last_dup_i = 0

  for i in range(vm_limit):
    #------------------------------------------------------------------------
    # The name we pass in will always be added to the vm_name array in the 
    # last available slot, unless there are no more slots.  
    #------------------------------------------------------------------------
    if vm_name[i] == "": # If empty slot, fill it and return index
      vm_name[i] = name

      #----------------------------------------------------------------------
      # If there are no others of the same name, there are no duplicates, so
      # call the tab name the name of the VM but set the tab_i value to 1, in
      # case a duplicate appears later.
      #----------------------------------------------------------------------
      if last_dup_i == 0:
        tab_name[i] = name
        tab_i[i] = 1

      else: # There was a duplicate, so use the number of duplicates as the unique id
        tab_name[i] = name + "/" + last_dup_i

      return i

    #------------------------------------------------------------------------
    # If we already have a VM of the same name as the one to "find", then we
    # must provide a different tab name for the original VM and any subsequent
    # ones of the same name.
    #------------------------------------------------------------------------
    if vm_name[i] == name: # If found name, force the tab name to be unique
      tab_name[i] = name + "/" + tab_i[i] # Use current tab_i value to name this

      tab_i[i] = tab_i[i] + 1 # There must be a duplicate name since name exists
      last_dup_i = tab_i[i] # Remember the duplicate id for next one

  msgbox_error("Exceeded maximum number of VMs: " + vm_limit)
  exit_program()

#---------------------------------------------------------------------------
# Retrieve the .vcl file from cssgate. Make sure you can ping the server
# first, so a more understandable error message can be provided to the
# user in case they don't have a good internet connection.
#---------------------------------------------------------------------------
def get_vcl_file():
  global server, ssh_options, cssgate_pw

  ping_result = ping(server)
  if ping_error != "":
    msgbox_info("Can't ping '" + server + "', error: " + ping_error)

  #--------------------------------------------------------------------------
  # Requires interaction, so do it this way vs. run_cmd()
  #
  # Still needs a click in the console window before entering password.
  #--------------------------------------------------------------------------
  ### print("\n*** Click in this console window prior to entering password.***\n")
  ### full_cmd = get_command_spec(' /c ') + cmd
  ### cmd_args = shlex.split(full_cmd)
  ### process = subprocess.run(cmd_args)

  if cssgate_pw is None:
    cssgate_pw = input_box("Please provide password for " + user + "@" + server + "; you may need to click inside box:", "", "*")

    if cssgate_pw is None: # If cancelled
      exit_program()

  if platform.system() == "Darwin":
    cmd = "SSHPASS='" + cssgate_pw + "'" + ' sshpass -e "' + re.sub(r'\\', r'\\\\', scp) + '" -P 22 -o StrictHostKeyChecking=no ' + ssh_options + ' ' + ' ' + user + "@" + server + ':"' + filename + '" "' + vcl_file + '"'
    write_log(-1, cmd)

    try:
      output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError  as e:
      output = e.output.decode('utf-8')
      if output.find("Permission denied") == 0:
        msgbox_error("Possible invalid password:\n\n" + cssgate_pw + "\n\n for "+ user + "@" + server)
        cssgate_pw = None 
        return False

      elif output.find(" Permission denied") != -1:
        msgbox_error(user + "@" + server + " cannot access the .vcl file called " + '"' + filename + '".' + "\n\n" + \
                     "Full error output:\n\n" + output)
        cssgate_pw = None
        return False

      else:
        msgbox_error(output)
        cssgate_pw = None
        return False
  
    for oline in output.splitlines():
      line = oline.decode('utf-8')
      if line.find(" denied") != -1:
        msgbox_error(user + "@" + server + " cannot access the .vcl file called " + '"' + filename + '".')
        cssgate_pw = None
        return False

    cssgate_pw = None
    return True

  if platform.system() == "Windows":
    cmd = '"' + scp + '" -P 22 ' + user + "@" + server + ":" + filename + ' "' + vcl_file + '"'
    write_log(-1, cmd)

    full_cmd = cmd

    cmd_args = shlex.split(full_cmd)

    process = popen_spawn.PopenSpawn(cmd_args)

    #------------------------------------------------------------------------
    # If this message, the host key is not saved for pscp or cygwin scp.
    #------------------------------------------------------------------------
    index = process.expect(['connection.', 'continue connecting', pexpect.TIMEOUT, pexpect.EOF], timeout=5)
    if index == 0:
      process.sendline("y")
      process.sendeof()
      process.kill(signal.CTRL_C_EVENT)
      return False

    elif index == 1:
      process.sendline("yes")
      process.sendeof()
      process.kill(signal.CTRL_C_EVENT)
      return False

    elif index == 3:
      process.sendeof()
      process.kill(signal.CTRL_C_EVENT)
      return False

    # Else Timeout

    process.sendline(cssgate_pw)
    process.sendeof()
  
    try:
      process.expect(['Cannot create file', 'lost connection'], timeout=2)
      msgbox_error("Access is denied to write to \"" + vcl_file + "\"")

    except pexpect.EOF:
      pass

    except pexpect.exceptions.TIMEOUT:
      msgbox_error("Possible invalid password:\n\n" + cssgate_pw + "\n\n for "+ user + "@" + server)

    process.wait()
  
    cssgate_pw = None

    return True
      
    '''
    process = subprocess.Popen(cmd_args, text=True, stdin=subprocess.PIPE,
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                               bufsize=1, universal_newlines=True)

    if not cssgate_pw is None:
      stderr_line = process.stderr.readline()
      print(f"stderr={stderr_line}")

      if stderr_line.find("not cached") >= 0:
        # Empty the stderr buffer for all but last line (can't find a way to
        # get that last line.

        stderr_line = process.stderr.readline()
        print(f"stderr={stderr_line}")

        while stderr_line.find("connection.") != 0:
          stderr_line = process.stderr.readline()
          print(f"stderr={stderr_line}")

        (stdout, stderr) = process.communicate(input="y\n")
        if stderr is not None:
          if stderr.find("Store key") >= 0:
            msgbox_info("Stored host key; re-trying transfer of VCL file.")
            return False
        return False

    else:
      (stdout, stderr) = process.communicate(input=cssgate_pw+"\n")

    if stderr is not None:
      if stderr.find("Access denied") != -1:
        msgbox_error("Possible invalid password:\n\n" + cssgate_pw + "\n\n for "+ user + "@" + server)
        cssgate_pw = None
        return False
      else:
        if len(stderr) != 0:
          msgbox_error(stderr)
          cssgate_pw = None
          return False

    cssgate_pw = None
    '''

    
    return True

##  if stdout is not None:
##    print(stdout)


#---------------------------------------------------------------------------
# Close the log, delete .rdp files, and exit the main script.
#---------------------------------------------------------------------------
def exit_program():
  global tk_root, tk_root_state
  write_log(-1, "--- Log ended ---")

  for vi in range(vm_limit):
    this_file = re.sub(r'\\', r'\\\\', os.path.join(get_profile_dir(), vm_name[vi] + "_" + str(vi)  + ".rdp"))
    if os.path.exists(this_file) == True:
      os.remove(this_file)

  if tk_root_state > 0:
    tk_root.destroy()

  sys.exit()

#---------------------------------------------------------------------------
# Look in the manage_vc file for the VCL file's location and other relevant
# info, and assign the values in that section to variables, then retain
# the information globally.
#---------------------------------------------------------------------------
def get_vcl_file_loc():
  global vcl_file, vcl_fn_user, server, user, filename
  user_info = [None for _ in range(3)]

  had_wildcard_user = False
  had_team_account = False

  string = "server,user,filename"
  vfl_var = string.split(",")

  #-------------------------------------------------------------------------
  # Get the vcl file location section from the .ini file.
  #-------------------------------------------------------------------------
  section = "vcl file location"

  if not section in ini_content:
    msgbox_error("Missing required section '" + section + "' in '" + ini_fn + "'")

    return False

  #-------------------------------------------------------------------------
  # Check for any missing variable values
  #-------------------------------------------------------------------------
  for var in vfl_var:
    if not var in ini_content[section]:
      msgbox_error("Missing variable '" + var + "' in section '" + section + "'")
      exit_program()

  #-------------------------------------------------------------------------
  # Assign this value to a script variable.
  #-------------------------------------------------------------------------
  server = ini_content[section]["server"]
  user = ini_content[section]["user"]
  filename = ini_content[section]["filename"]

  vcl_fn_user = user.lower()

  #-------------------------------------------------------------------------
  # If there are command line args, use them in preference to .ini values.
  #-------------------------------------------------------------------------
  if len(sys.argv) > 1: # If passed in user information of form: root user
    user = sys.argv[1].lower() # Override .ini file settings
    orig_user = user
    vcl_fn_user = sys.argv[2].lower()

  else:
    #-------------------------------------------------------------------------
    # Else no command line args, so work with what was read from .ini file.
    #-------------------------------------------------------------------------
    if user == "*":  # If a wildcard
      had_wildcard_user = True
      user = getpass.getuser() # Get a username from the system

    #-------------------------------------------------------------------------
    # Always prompt to accept or change, in case they hardcoded a username in
    # the .ini file and they now need to specify a team account as well.
    #-------------------------------------------------------------------------
    user = input_box("Please accept or change the user name:", user, None)

    if user is None: # If CANCEL from input_box???
      exit_program()

    user = user.lower()

    #-------------------------------------------------------------------------
    # User info could be just the user name, or "username teamacct" (or also
    # "root username").
    #-------------------------------------------------------------------------
    user_info = user.split(" ")

    if len(user_info) == 2: # If two strings
      had_team_account = True
      user = user_info[0].lower()
      vcl_fn_user = user_info[1].lower()

    else:
      #----------------------------------------------------------------------
      # A common error is for students to not use a space between user name 
      # and team account name.
      #
      # Check for missing blank between user name and team name starting 
      # with an underscore
      #----------------------------------------------------------------------
      underscore_pos = user.find("_")

      if underscore_pos >= 1: # If someone concatenated user and vcl_fn_user
        user1 = user[:underscore_pos].lower()
        vcl_fn_user = user[underscore_pos+1:].lower()

        msgbox_error('There should not be an underscore in the middle of the user name "' + user + '", but a space between your UW Net Id and the team name (e.g., "' + user1 + " " + vcl_fn_user + '"). If you forgot the space, restart manage_vc and add the space between your UW Net Id and team name.')
        user = user1

        exit_program()
  
      vcl_fn_user = user.lower()

      pos = user.find("@")
      if pos > -1:
        good_user = user[0:pos]
        msgbox_error('The user id must be a valid UW Net Id. Email addresses contain "@". The part BEFORE the "@" could be your UW Net Id; in this case, it is likely "' + good_user + '". Please restart manage_vc and provide just your UW Net Id for a user name.')
        exit_program()

  write_log(-1, "user info: " + user + ' ' + vcl_fn_user)

  #-------------------------------------------------------------------------
  # If there is an absolute path, then the user name is the second to
  # last part of the path.
  #-------------------------------------------------------------------------
  if filename.find("/") == 0:  # If absolute path
    matches = re.match("(\w+?)\/[^/]+?", filename)
    vcl_fn_user = matches.group(1)

  elif (had_wildcard_user == True) or (had_team_account == True):
    filename = "/home/INSTTECH/" + vcl_fn_user + "/" + filename

  #-------------------------------------------------------------------------
  # The VCL file will be placed here after retrieving from cssgate.
  #-------------------------------------------------------------------------
  vcl_file = os.path.join(destination_dir,  vcl_fn + "." + vcl_fn_user)

  return True

#-------------------------------------------------------------------------
# Parse the desktop section of the .ini file similarly to others.
#-------------------------------------------------------------------------
def get_desktop_section():
  global vcl_desktop_width, vcl_desktop_height, vcl_desktop_bpp, vcl_desktop_mode
  
  section = "desktop"
  if not section in ini_content:
    msgbox_error("Missing required section '" + section + "' in '" + ini_fn + "'")

    return False

  string = "width,height,bpp,mode"
  desktop_var = string.split(",")

  #-------------------------------------------------------------------------
  # Assign value to a script variable.
  #-------------------------------------------------------------------------
  if "width" in ini_content[section]:
    vcl_desktop_width = ini_content[section]["width"]

  if "height" in ini_content[section]:
    vcl_desktop_height = ini_content[section]["height"]

  if "bpp" in ini_content[section]:
    vcl_desktop_bpp = ini_content[section]["bpp"]

  if "mode" in ini_content[section]:
    vcl_desktop_mode = ini_content[section]["mode"]

  write_log(-1, "desktop info: width=" + vcl_desktop_width + ' height=' + vcl_desktop_height + ' bpp=' + vcl_desktop_bpp + ' mode=' + vcl_desktop_mode)

#-------------------------------------------------------------------------
# Parse the ssh info section similarly to other sections.
#-------------------------------------------------------------------------
def get_ssh():
  #-------------------------------------------------------------------------
  # Ignore what user specifies for scp and ssh and use what is in ScriptDir
  # only
  #-------------------------------------------------------------------------
  global scp, ssh
  scp = "scp"
  ssh = "ssh"

  if platform.system() == "Windows":
    scp = ""
    orig_scp = scp

    ssh = ""
    orig_ssh = ssh

    if scp == "":
#      scp = re.sub(r'\\', r'\\\\', get_script_dir() + "\pscp.exe")
#      scp = re.sub(r'\\pscp', r'\\\\pscp', os.path.join(get_script_dir(), "pscp.exe"))
      scp =  os.path.join(get_script_dir(), "pscp.exe")

    if ssh == "":
#      ssh = re.sub(r'\\', r'\\\\', get_script_dir() + "\plink.exe")
#      ssh = re.sub(r'\\plink', r'\\\\plink', os.path.join(get_script_dir(), "plink.exe"))
      ssh = os.path.join(get_script_dir(), "plink.exe")

    if os.path.exists(scp) == False:
      scp = "C:\Program Files (x86)\PuTTY\pscp.exe"

    if os.path.exists(scp) == False:
#      scp = re.sub(r'\\', r'\\\\', get_script_dir() + "\pscp.exe")
      scp =  re.sub(r'\\', r'\\\\', os.path.join(get_script_dir(), "pscp.exe"))

      if os.path.exists(scp) == False:
        msgbox_error("Can't find scp program '" + orig_scp + "'")
        exit_program()

    if os.path.exists(ssh) == False:
      ssh = "C:\Program Files (x86)\PuTTY\plink.exe"

    if os.path.exists(ssh) == False:
#      ssh = re.sub(r'\\', r'\\\\', get_script_dir() + "\plink.exe")
      ssh =  re.sub(r'\\', r'\\\\', os.path.join(get_script_dir(), "plink.exe"))

      if os.path.exists(ssh) == False:
        msgbox_error("Can't find ssh program '" + orig_ssh + "'")
        exit_program()

#---------------------------------------------------------------------------
# Hash the password from the VCL file using the crypt executablecinto a 
# Windows hashed passwordso it can be used as part of the mstsc RDP spec.
#---------------------------------------------------------------------------
def hash_pw(password):
  if platform.system() == "Windows":
    pw_hash_fn = re.sub(r'\\', r'\\\\', os.path.join(get_profile_dir(), "pw_hash"))

    if os.path.exists(pw_hash_fn) == True:
      os.remove(pw_hash_fn)

    cmd = '"' + crypt_exec + '" > "' + pw_hash_fn + '" ' + password
    write_log(-1, cmd)

    run_cmd(get_command_spec(' /d /c ') + quote(cmd), False, False)
  
    pw_hash_handle = open(pw_hash_fn, 'r')
    hashed_pw = pw_hash_handle.readline()[0:-1]
    pw_hash_handle.close()

    os.remove(pw_hash_fn)
  
    return hashed_pw

  else: # Any non-Windows platform won't have Windows credentials used by cryptrdp5, so it will fail
    return '???'

#---------------------------------------------------------------------------
# Create the RDP file for mstsc using the information extracted from
# various files. Many features are disabled to prevent users from
# easily overwhelming the network via USB drive file transfers, printing
# and whatnot.
#---------------------------------------------------------------------------
def write_rdp_file(rdp_fn, host_and_port, domain, user, pw_hash):
  rdp_file = open(rdp_fn, 'w')
  rdp_file.write("screen mode id:i:" + vcl_desktop_mode + "\n")
  rdp_file.write("desktopwidth:i:" + vcl_desktop_width + "\n")
  rdp_file.write("desktopheight:i:" + vcl_desktop_height + "\n")
  rdp_file.write("session bpp:i:" + vcl_desktop_bpp + "\n")
  rdp_file.write("winposstr:s:0,1,32,68,800,572" + "\n")
  rdp_file.write("full address:s:" + host_and_port + "\n")
  rdp_file.write("compression:i:1" + "\n")
  rdp_file.write("keyboardhook:i:2" + "\n")
  rdp_file.write("audiomode:i:2" + "\n")
  rdp_file.write("redirectdrives:i:0" + "\n")
  rdp_file.write("redirectprinters:i:0" + "\n")
  rdp_file.write("redirectcomports:i:0" + "\n")
  rdp_file.write("redirectsmartcards:i:0" + "\n")
  rdp_file.write("displayconnectionbar:i:1" + "\n")
  rdp_file.write("autoreconnection enabled:i:1" + "\n")
  rdp_file.write("authentication level:i:0" + "\n")
  rdp_file.write("username:s:" + user + "\n")
  rdp_file.write("domain:s:" + domain + "\n")
  rdp_file.write("alternate shell:s:" + "\n")
  rdp_file.write("shell working directory:s:" + "\n")
  if pw_hash != '':
    rdp_file.write("password 51:b:" + pw_hash + "\n")
  rdp_file.write("disable wallpaper:i:1" + "\n")
  rdp_file.write("disable full window drag:i:0" + "\n")
  rdp_file.write("disable menu anims:i:0" + "\n")
  rdp_file.write("disable themes:i:0" + "\n")
  rdp_file.write("disable cursor setting:i:0" + "\n")
  rdp_file.write("bitmapcachepersistenable:i:1" + "\n")
  rdp_file.close()

#---------------------------------------------------------------------------
# Keep a record of what ssh/scp versions used in case of problems.
#---------------------------------------------------------------------------
def log_ssh_versions():
  temp_file = os.path.join(get_profile_dir(), "manage_vc.out")

  run_cmd(get_command_spec(' /c ') + quote(ssh + ' -V > "' + temp_file + '" 2>&1'), False, False)

  with open(temp_file, 'r') as temp_handle:
    line = temp_handle.readline()
    line.rstrip() # Remove trailing newlines and whitespace

  write_log(-1, "ssh: " + ssh)
  write_log(-1, "ssh: " + line)

  if platform.system() == "Windows": # If windows
    if scp.find('pscp') > 0: # If pscp -- scp doesn't have any option for version
      run_cmd(get_command_spec(' /c ') + quote(scp + '" >"' + temp_file) + ' -P 22 -V', False, False)

      temp_handle = open(temp_file, 'r')
      line = temp_handle.readline()
      line.rstrip() # Remove trailing newlines and whitespace
      temp_handle.close()

      write_log(-1, "scp: " + scp)
      write_log(-1, "scp: " + line)

  os.remove(temp_file)

#---------------------------------------------------------------------------
# Read the VM's NIC info to determine what kind of networking methods
# are usable for guest RDP, and possibly some guest IP information.
#---------------------------------------------------------------------------
def get_nic_info(vi):
  global nic_info

  #-------------------------------------------------------------------------
  # Clear all previous info for this VM's nic_info
  #-------------------------------------------------------------------------
  for i in range(nic_limit):
    for j in range(nic_info_limit):
      nic_info[vi][i][j] = ""

  lines = send_vcl_command(vi, auth + "show_nic " + vm_name[vi], 1, False)

  for oline in lines.splitlines():
    line = oline.decode('utf-8')
    line.rstrip() # Remove trailing newlines and whitespace

    #-----------------------------------------------------------------------
    # NIC information from show_nic can roughly be divided by colons
    # NIC 1:                       MAC: 080027B19F45, Attachment: Internal Network 'srondeau_intnet', Cable connected: on, Trace: off (file: none), Type: 82540EM, Reported speed: 0 Mbps, Boot priority: 0, Promisc Policy: deny, Bandwidth group: none
    #-----------------------------------------------------------------------
    value = line.split(":")

    #-----------------------------------------------------------------------
    # The first colon-delimited value contains the NIC number as the second
    # space-delimited string; e.g.
    # NIC 1
    #-----------------------------------------------------------------------
    subval = value[0].split(" ")

    #-----------------------------------------------------------------------
    # Remember the 0-based NIC number value to index into the array of NIC info
    # per VM
    #-----------------------------------------------------------------------
    nicnum = int(subval[1])-1

    #-----------------------------------------------------------------------
    # However, there might not be any information defined for a NICi,
    # so we look for "disabled" within the second colon-delimited string; e.g.
    # NIC 4:                       disabled
    # If disabled is found, move on to next NIC line
    #-----------------------------------------------------------------------
    pos = value[1].find("disabled")

    if pos >= 0:
      continue

    #-----------------------------------------------------------------------
    # If something after the NIC number
    #-----------------------------------------------------------------------
    if len(subval) > 2:
      #---------------------------------------------------------------------
      # If settings for NIC, move on to next NIC line; e.g.
      # NIC 2 Settings:  MTU: 0, Socket (send: 64, receive: 64), TCP Window (send:64, receive: 64)
      #---------------------------------------------------------------------
      pos = subval[2].find("Settings")
      if pos >= 0:
        continue

      #---------------------------------------------------------------------
      # If Rule for NIC, retain the first containing "ms-wbt-server" and
      # discard the rest.
      # NIC 2 Rule(0):   name = guestms-wbt-server, protocol = tcp, host ip = , host port = 51515, guest ip = , guest port = 3389
      #---------------------------------------------------------------------
      pos = subval[2].find("Rule")
      if pos >= 0:
        pos = value[1].find("ms-wbt-server")
        if pos >= 0:
          nic_info[vi][nicnum][ni_rule] = value[1][:len(value[1])-1]

        continue

    #---------------------------------------------------------------------
    # Retain network attachment method, minus initial space character
    #---------------------------------------------------------------------
    subval = value[3].split(",")
    nic_info[vi][nicnum][ni_method] = subval[0][1:] 

    #---------------------------------------------------------------------
    # Retain network "Cable connected" status: on or off, after removing
    # initial space character.
    #---------------------------------------------------------------------
    subval = value[4].split(",")
    nic_info[vi][nicnum][ni_cc_status] = subval[0][1:]

#---------------------------------------------------------------------------
# Extract the VM guest OS's IP information
#---------------------------------------------------------------------------
def get_vm_ip(vi, ni):
  #---------------------------------------------------------------------------
  # show_vm_ip uses one-based NIC numbers
  #---------------------------------------------------------------------------
  nicnum = ni+1
  lines = send_vcl_command(vi, auth + "show_vm_ip " + vm_name[vi] + " " + nicnum, 1, False)

  for oline in lines.splitlines():
    line = oline.decode('utf-8')
    line.rstrip() # Remove trailing newlines and whitespace
    break # Only first line

  word = line.split(" ")

  #---------------------------------------------------------------------------
  # If "The vm_name has not been started" or  "No value set!"
  #---------------------------------------------------------------------------
  if (word[0] == "No") or (word[0] == "The"):
    return ""

  return word[1]

#-----------------------------------------------------------------------------
# progress functions
#-----------------------------------------------------------------------------
def progress_on(title):
  global tk_root, progress, max_width, max_height, pb_top, progress_label

  if pb_top is not None:
    return False

  width = max_width//4
  pb_top = Toplevel(tk_root)
  pb_top.title(title)
  pb_top.geometry(str(width*3) + 'x80')

  pb_top.focus_set()
  pb_top.grab_set()
  pb_top.overrideredirect()
  pb_top.resizable(False, False)
  #----------------------------------------------------------------------------
  # In case user won't click on either button, they can close window and do 
  # nothing.
  #----------------------------------------------------------------------------
  pb_top.protocol("WM_DELETE_WINDOW", on_pb_close) 

  progress = ttk.Progressbar(pb_top, orient='horizontal', mode='determinate', length=width*2)
  progress["maximum"] = 100.0

  progress.grid(column=0, row=0, columnspan=2, padx=10, pady=10)

  progress_label = StringVar(pb_top, "")
 
  ttk.Label(pb_top, textvariable=progress_label).grid(column=0, row=1, columnspan=2, padx=10, pady=10)

  pb_top.after(500, progress_update)
  progress_set(0.0)

#---------------------------------------------------------------------------
# Update progress
#---------------------------------------------------------------------------
def progress_update():
  global pb_top
  pb_top.update()

#---------------------------------------------------------------------------
# Set progress
#---------------------------------------------------------------------------
def progress_set(value):
  global pb_top, progress, progress_label
  progress['value'] = value
  progress_label.set(f"Current progress: {value}%")

#---------------------------------------------------------------------------
# Close Progress Bar window when 100% complete
#---------------------------------------------------------------------------
def progress_off():
  global pb_top, progress_label
  pb_top.destroy()
  progress_label = None
  pb_top = None

#---------------------------------------------------------------------------
# When progress bar window is closed
#---------------------------------------------------------------------------
def on_pb_close():
  pb_top.destroy()
    
#---------------------------------------------------------------------------
# Configure RDP based on the NIC information extracted.
#---------------------------------------------------------------------------
def setup_rdp_from_nic_info(vi):
  global rdp_info

  #---------------------------------------------------------------------------
  # Clear out old values
  #---------------------------------------------------------------------------
  for i in range(rdp_info_limit):
    rdp_info[vi][i] = ""

  found_rdp_info = 0

  if platform.system() == "Windows":
    progress_on("Gathering recent network info...")

  get_nic_info(vi)

  if platform.system() == "Windows":
    progress_set(50.0)

  #---------------------------------------------------------------------------
  # Use information from the last NAT/Bridged NIC that has valid info
  #---------------------------------------------------------------------------
  for ni in range(nic_limit):
    #------------------------------------------------------------------------
    # If a NIC is off, go to next
    #------------------------------------------------------------------------
    if (nic_info[vi][ni][ni_cc_status] == 'off') or (nic_info[vi][ni][ni_cc_status] == ''):
      continue

    #------------------------------------------------------------------------
    # If NAT, look for rule saved in index 3.
    #------------------------------------------------------------------------
    if nic_info[vi][ni][ni_method].find("NAT") >= 0:
      #----------------------------------------------------------------------
      # If relevant rule info, save the RDP host port.
      #----------------------------------------------------------------------
      
      value = nic_info[vi][ni][ni_rule].split(", ")

      if len(value) >= 4: # If NAT port mapping info present
        subval = value[3].split("= ")

        found_rdp_info = 1
        rdp_info[vi][ri_host] = host_addr[vi]
        rdp_info[vi][ri_port] = subval[1]

    elif nic_info[vi][ni][ni_method].find("Bridged") >= 0:
      host = get_vm_ip(vi, ni)

      #----------------------------------------------------------------------
      # If IP address returned, save it and use default 3389 port
      # Else no IP address due to not-running VM or no NIC info for NIC number
      #----------------------------------------------------------------------
      if host != "":
        found_rdp_info = 1
        rdp_info[vi][ri_host] = host
        rdp_info[vi][ri_port] = "3389"

  if platform.system() == "Windows":
    progress_set(100.0)
    time.sleep(0.5)
    progress_off()

  return found_rdp_info

#---------------------------------------------------------------------------
# Determine if VM is running by parsing its state.
#---------------------------------------------------------------------------
def is_running(vi):
  lines = send_vcl_command(vi, "vcl_status " + vm_name[vi], 1, False)

  for oline in lines.splitlines():
    line = oline.decode('utf-8')
    line.rstrip() # Remove trailing newlines and whitespace

    if (line.find("State:") >= 0) and \
       (line.find("running") >= 0):
      return 1

  return 0

#---------------------------------------------------------------------------
# To avoid some problems with plink/pscp, make sure there are host keys
# for each host saved in the registry.
#---------------------------------------------------------------------------
def have_host_key(key):
  found=0

  if ssh.find("plink") == -1: # If not plink, don't check plink-specific registry entry
    return found

  if platform.system != "Windows":
    return found

  maxvals=1000

  access_registry = winreg.ConnectRegistry(None,winreg.HKEY_CURRENT_USER)

  access_key = winreg.OpenKey(access_registry,"Software\\SimonTatham\\PuTTY\\SshHostKeys")
  for i in range(maxvals):
    try:
      var = winreg.EnumValue(access_key, i+1)

    except OSError:
      break

    match = re.match(key + "", var)
    if match:
      found = 1
      break

  return found

#----------------------------------------------------------------------------
# ping a host
#----------------------------------------------------------------------------
def ping(host):
  global ping_error

  ping_error = ""

  temp_file = re.sub(r'\\', r'\\\\', os.path.join(get_profile_dir(), "ping.out"))

  if platform.system() == "Windows":
    run_cmd('ping >"' + temp_file + '" -w 1 -n 1 ' + host, False, False)

    temp_handle = open(temp_file, 'r')
    line = temp_handle.readline()
    line = temp_handle.readline()
    line = temp_handle.readline()
    line.rstrip() # Remove trailing newlines and whitespace
    temp_handle.close()

  else: # non-windows
    if platform.system() == "Darwin":
      run_cmd("ping >" + temp_file + " -c 1 " + host, False, False)
    else:
      run_cmd("ping >" + temp_file + " -W 1 -c 1 " + host, False, False)

    temp_handle = open(temp_file, 'r')
    line = temp_handle.readline()
    line = temp_handle.readline() # Second line has reply results
    line.rstrip() # Remove trailing newlines and whitespace
    temp_handle.close()

    if line.find("64 bytes from") >= 0:
      os.remove(temp_file)
      return True

  if line.find(" from ") >= 0:
    os.remove(temp_file)
    return True

  ping_error = line
  os.remove(temp_file)
  
  return False

#----------------------------------------------------------------------------
# Optionally quote a command
#----------------------------------------------------------------------------
def quote(cmd):
  if platform.system() == "Windows":
    return '"' + cmd + '"'
  return cmd

#----------------------------------------------------------------------------
# Get the command shell for Windows
#----------------------------------------------------------------------------
def get_command_spec(options):
  if platform.system() == "Windows":
    return 'cmd' + options
  return ""

#---------------------------------------------------------------------------
# Main
#---------------------------------------------------------------------------
def main():
  global tk_root, tk_root_state
  tk_root_state = 2

  populate_gui()

  tk_root.mainloop() # Process window events

#---------------------------------------------------------------------------
# We must find the index of the VM information corresponding to the host
# IP address, as the only way to manage the VM is by sending commands
# to the host on which it resides.
#---------------------------------------------------------------------------
def find_vm_index(addr):
  for i in range(vmi+1):
    if addr == host_addr[i]:
      return i
  
  return -1

#---------------------------------------------------------------------------
# Find the index of the VM associated with the control id of the GUI event.
#---------------------------------------------------------------------------
def get_vi(array):
  return vms.index('current')

#-----------------------------------------------------------------------------
# Get directory where script was found
#-----------------------------------------------------------------------------
def get_script_dir():
  ## return os.path.dirname(os.path.abspath(__file__))
  ## return os.getcwd()
  if getattr(sys, 'frozen', False):
    return os.path.dirname(os.sys.executable)
  else:
    return os.path.dirname(os.path.abspath(__file__))

#-----------------------------------------------------------------------------
# Get CRLF
#-----------------------------------------------------------------------------
def get_crlf():
  return "\n"

tk_root_state = 0

create_gui() # For message boxes, prior to getting VM information

crypt_exec = re.sub(r'\\', r'\\\\', os.path.join(get_script_dir(), "cryptRDP5.exe"))

#------------------------------------------------------------------------------
# Get width of button name
#------------------------------------------------------------------------------
def get_button_width(name, win_adj, mac_adj, lin_adj):
  global font_width

  width = len(name)
  if platform.system() == "Windows":
    width = width + win_adj

  elif platform.system() == "Darwin":
    width = width*font_width + mac_adj

  else:
    width = width + lin_adj

  return width

#------------------------------------------------------------------------------
# Get width of list box
#------------------------------------------------------------------------------
def get_listbox_width(init_width, win_adj, mac_adj, lin_adj):
  width = init_width

  if platform.system() == "Windows":
    width = width + win_adj

  elif platform.system() == "Darwin":
    width = width + mac_adj

  else:
    width = width + lin_adj

  return width

#------------------------------------------------------------------------------
# Define destination for files written or copied.
#------------------------------------------------------------------------------

destination_dir = re.sub(r'\\', r'\\\\', get_profile_dir())

#------------------------------------------------------------------------------
# Initialize the next port indices.
#------------------------------------------------------------------------------
for i in range(vm_limit):
  next_port_index[i] = -1

#------------------------------------------------------------------------------
# Open a file for log entries.
#------------------------------------------------------------------------------
log_file = open(os.path.join(destination_dir, log_fn), mode='w', encoding="utf-8")

write_log(-1, "--- Log started for manage_vc " + mv_version + " ---");

#------------------------------------------------------------------------------
# Get information from manage_vc.ini file, which could be overridden by
# either the command line or user prompts.
#------------------------------------------------------------------------------
if os.path.exists(ini_fn) == False:
  msgbox_error("Can't find '" + ini_fn + "'")
  exit_program()

ini_content = ConfigParser()
ini_content.read(ini_fn)

write_log(-1, "ScriptDir: " + get_script_dir())

get_ssh()

log_ssh_versions()

if get_vcl_file_loc() == 0:
  exit_program()

get_desktop_section()

#------------------------------------------------------------------------------
# If there are no command line arguments, determine where to get the .vcl
# file. If the file exists, ask the user if it should be reused, which saves
# time and avoids retrieving it from cssgate. On the other hand, the .vcl file
# could have been updated and the user informed, or it is the beginning
# of a new quarter when most student .vcl files are new versions, and hence 
# they need to retrieve it from cssgate.
#------------------------------------------------------------------------------
reuse = 7

if len(sys.argv) == 1: # If no command args
  if os.path.exists(vcl_file) == True: 
    msgbox_yesnocancel("'" + re.sub(r'\\\\', r'\\', vcl_file) + "' already exists. Reuse (Yes) or get latest version (No)?")

    if msgbox_cancel == True:
      exit_program()

    if msgbox_yes == True:
      reuse = 0

  else: # File does not exist; get latest version
    reuse = 7

  if reuse == 7: # Get latest version
    while get_vcl_file() == False:
      pass

#------------------------------------------------------------------------------
# If after trying to retrieve the .vcl file, it still isn't present, there
# must have been some error, and we cannot continue without it.
#------------------------------------------------------------------------------
if os.path.exists(vcl_file) == False: 
  msgbox_error("Can't find '" + vcl_file + "'")
  exit_program()

#------------------------------------------------------------------------------
# If the file is there but unreadable, something must have happened to
# make it that way, so warn the user, delete the file (if possible) and
# have them try again, perhaps after consulting the troubleshooting section.
#------------------------------------------------------------------------------
if os.path.getsize(vcl_file) == 0:
  msgbox_error("Access is not allowed for '" + vcl_file + "'")
  os.remove(vcl_file)
  exit_program()

#------------------------------------------------------------------------------
# Read and parse lines from the .vcl file, which establishes all of
# the information needed to send commands to the host(s) to manage the
# VMs.
#------------------------------------------------------------------------------
read_vcl_file()

#------------------------------------------------------------------------------
# Since different implementations of ssh and scp are allowed (primarily
# due to the lack of ssh/scp in Windows prior to Windows 10 1809), we
# may need to set up alternate ssh/scp command options to work properly.
#------------------------------------------------------------------------------
setup_send_vcl_command()

## show_vcl_info()
top_pos = 5

#-------------------------------------------------------------------------
# Set up the max width and height of a tab, allowing multiple rows of
# tabs if need be.
#-------------------------------------------------------------------------
max_width = main_window_width - 20
max_height = main_window_height - 10

if platform.system() == "Windows":
  import winreg

#------------------------------------------------------------------------------
# Create the GUI from all of the known VCL information. The idea is
# one tab per VM, without highlighting the host on which the VM resides
# since it doesn't really matter to the user.
#------------------------------------------------------------------------------
main()
